# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Collaborator module."""

import logging
from enum import Enum
from time import sleep
from typing import List, Optional

import openfl.callbacks as callbacks_module
from openfl.databases import TensorDB
from openfl.pipelines import NoCompressionPipeline, TensorCodec
from openfl.protocols import utils
from openfl.transport.grpc.aggregator_client import AggregatorGRPCClient
from openfl.utilities import TensorKey

logger = logging.getLogger(__name__)


class OptTreatment(Enum):
    """Optimizer Methods.

    Attributes:
        RESET (int): Resets the optimizer state at the beginning of each round.
        CONTINUE_LOCAL (int): Continues with the local optimizer state from
            the previous round.
        CONTINUE_GLOBAL (int): Continues with the federally averaged optimizer
            state from the previous round.
    """

    RESET = 1
    CONTINUE_LOCAL = 2
    CONTINUE_GLOBAL = 3


class Collaborator:
    r"""The Collaborator object class.

    Attributes:
        collaborator_name (str): The common name for the collaborator.
        aggregator_uuid (str): The unique id for the client.
        federation_uuid (str): The unique id for the federation.
        client (object): The client object.
        task_runner (object): The task runner object.
        task_config (dict): The task configuration.
        opt_treatment (str)*: The optimizer state treatment.
        device_assignment_policy (str): [Deprecated] The device assignment policy.
        use_delta_updates (bool)*: If True, only model delta gets sent. If False,
            whole model gets sent to collaborator.
        compression_pipeline (object): The compression pipeline.
        db_store_rounds (int): The number of rounds to store in the database.
        single_col_cert_common_name (str): The common name for the single
            column certificate.

    .. note::
        \* - Plan setting.
    """

    def __init__(
        self,
        collaborator_name,
        aggregator_uuid,
        federation_uuid,
        client: AggregatorGRPCClient,
        task_runner,
        task_config,
        opt_treatment="RESET",
        device_assignment_policy="CPU_ONLY",
        use_delta_updates=False,
        compression_pipeline=None,
        db_store_rounds=1,
        log_memory_usage=False,
        write_logs=False,
        callbacks: Optional[List] = [],
        secure_aggregation=False,
    ):
        """Initialize the Collaborator object.

        Args:
            collaborator_name (str): The common name for the collaborator.
            aggregator_uuid (str): The unique id for the client.
            federation_uuid (str): The unique id for the federation.
            client (object): The client object.
            task_runner (object): The task runner object.
            task_config (dict): The task configuration.
            opt_treatment (str, optional): The optimizer state treatment.
                Defaults to 'RESET'.
            device_assignment_policy (str, optional): The device assignment
                policy. Defaults to 'CPU_ONLY'.
            use_delta_updates (bool, optional): If True, only model delta gets
                sent. If False, whole model gets sent to collaborator.
                Defaults to False.
            compression_pipeline (object, optional): The compression pipeline.
                Defaults to None.
            db_store_rounds (int, optional): The number of rounds to store in
                the database. Defaults to 1.
            callbacks (list, optional): List of callbacks. Defaults to None.
        """
        # for protobuf compatibility we would really want this as an object
        self.single_col_cert_common_name = ""

        self.collaborator_name = collaborator_name
        self.aggregator_uuid = aggregator_uuid
        self.federation_uuid = federation_uuid

        self.compression_pipeline = compression_pipeline or NoCompressionPipeline()
        self.tensor_codec = TensorCodec(self.compression_pipeline)
        self.tensor_db = TensorDB()
        self.db_store_rounds = db_store_rounds

        self.task_runner = task_runner
        self.use_delta_updates = use_delta_updates

        self.client = client

        self.task_config = task_config

        # RESET/CONTINUE_LOCAL/CONTINUE_GLOBAL
        if hasattr(OptTreatment, opt_treatment):
            self.opt_treatment = OptTreatment[opt_treatment]
        else:
            logger.error("Unknown opt_treatment: %s.", opt_treatment.name)
            raise NotImplementedError(f"Unknown opt_treatment: {opt_treatment}.")
        self.task_runner.set_optimizer_treatment(self.opt_treatment.name)

        logger.warning(
            "Argument `device_assignment_policy` is deprecated and will be removed in the future."
        )
        del device_assignment_policy

        # Secure aggregation
        self._secure_aggregation_enabled = secure_aggregation
        if self._secure_aggregation_enabled:
            self._private_mask = None
            self._shared_mask = None
            secure_aggregation_callback = callbacks_module.SecAggBootstrapping()
            if isinstance(callbacks, callbacks_module.Callback):
                callbacks = [callbacks, secure_aggregation_callback]
            elif isinstance(callbacks, list):
                callbacks.append(secure_aggregation_callback)
            else:
                callbacks = [secure_aggregation_callback]

        # Callbacks
        self.callbacks = callbacks_module.CallbackList(
            callbacks,
            add_memory_profiler=log_memory_usage,
            add_metric_writer=write_logs,
            tensor_db=self.tensor_db,
            origin=self.collaborator_name,
            client=self.client,
        )

    def ping(self):
        """Ping the Aggregator."""
        self.client.ping()

    def run(self):
        """Run the collaborator."""
        # Experiment begin
        self.callbacks.on_experiment_begin()

        while True:
            tasks, round_num, sleep_time, time_to_quit = self.client.get_tasks()

            if time_to_quit:
                break

            if not tasks:
                sleep(sleep_time)
                continue

            # Round begin
            logger.info("Received Tasks: %s", tasks)
            self.callbacks.on_round_begin(round_num)

            # Run tasks
            logs = {}
            for task in tasks:
                metrics = self.do_task(task, round_num)
                logs.update(metrics)

            # Round end
            self.tensor_db.clean_up(self.db_store_rounds)
            self.callbacks.on_round_end(round_num, logs)

        # Experiment end
        self.callbacks.on_experiment_end()
        logger.info("Received shutdown signal. Exiting...")

    def do_task(self, task, round_number) -> dict:
        """Perform the specified task.

        Args:
            task (list_of_str): List of tasks.
            round_number (int): Actual round number.

        Returns:
            A dictionary of reportable metrics of the current collaborator for the task.
        """
        # map this task to an actual function name and kwargs
        if isinstance(task, str):
            task_name = task
        else:
            task_name = task.name
        func_name = self.task_config[task_name]["function"]
        kwargs = self.task_config[task_name]["kwargs"]

        # this would return a list of what tensors we require as TensorKeys
        required_tensorkeys_relative = self.task_runner.get_required_tensorkeys_for_function(
            func_name, **kwargs
        )

        # models actually return "relative" tensorkeys of (name, LOCAL|GLOBAL,
        # round_offset)
        # so we need to update these keys to their "absolute values"
        required_tensorkeys = []
        for (
            tname,
            origin,
            rnd_num,
            report,
            tags,
        ) in required_tensorkeys_relative:
            if origin == "GLOBAL":
                origin = self.aggregator_uuid
            else:
                origin = self.collaborator_name

            # rnd_num is the relative round. So if rnd_num is -1, get the
            # tensor from the previous round
            required_tensorkeys.append(
                TensorKey(tname, origin, rnd_num + round_number, report, tags)
            )

        # print('Required tensorkeys = {}'.format(
        # [tk[0] for tk in required_tensorkeys]))
        input_tensor_dict = {
            k.tensor_name: self.get_data_for_tensorkey(k) for k in required_tensorkeys
        }

        # now we have whatever the model needs to do the task
        # Tasks are defined as methods of TaskRunner
        func = getattr(self.task_runner, func_name)
        logger.debug("Using TaskRunner subclassing API")

        global_output_tensor_dict, local_output_tensor_dict = func(
            col_name=self.collaborator_name,
            round_num=round_number,
            input_tensor_dict=input_tensor_dict,
            **kwargs,
        )
        # If secure aggregation is enabled, add masks to the dict to be shared
        # with the aggregator.
        if self._secure_aggregation_enabled:
            self._apply_masks(global_output_tensor_dict)

        # Save global and local output_tensor_dicts to TensorDB
        self.tensor_db.cache_tensor(global_output_tensor_dict)
        self.tensor_db.cache_tensor(local_output_tensor_dict)

        # send the results for this tasks; delta and compression will occur in
        # this function
        metrics = self.send_task_results(global_output_tensor_dict, round_number, task_name)

        return metrics

    def get_data_for_tensorkey(self, tensor_key):
        """Resolve the tensor corresponding to the requested tensorkey.

        Args:
            tensor_key (namedtuple): Tensorkey that will be resolved locally or
            remotely. May be the product of other tensors.

        Returns:
            nparray: The decompressed tensor associated with the requested
                tensor key.
        """
        # try to get from the store
        tensor_name, origin, round_number, report, tags = tensor_key
        logger.debug("Attempting to retrieve tensor %s from local store", tensor_key)
        nparray = self.tensor_db.get_tensor_from_cache(tensor_key)

        # if None and origin is our client, request it from the client
        if nparray is None:
            if origin == self.collaborator_name:
                logger.info(
                    f"Attempting to find locally stored {tensor_name} tensor from prior round..."
                )
                prior_round = round_number - 1
                while prior_round >= 0:
                    nparray = self.tensor_db.get_tensor_from_cache(
                        TensorKey(tensor_name, origin, prior_round, report, tags)
                    )
                    if nparray is not None:
                        logger.debug(
                            f"Found tensor {tensor_name} in local TensorDB for round {prior_round}"
                        )
                        return nparray
                    prior_round -= 1
                logger.info(f"Cannot find any prior version of tensor {tensor_name} locally...")
            # Determine whether there are additional compression related
            # dependencies.
            # Typically, dependencies are only relevant to model layers
            tensor_dependencies = self.tensor_codec.find_dependencies(
                tensor_key, self.use_delta_updates
            )
            logger.debug(
                "Unable to get tensor from local store..."
                "attempting to retrieve from client len tensor_dependencies"
                f" tensor_key {tensor_key}"
            )
            if len(tensor_dependencies) > 0:
                # Resolve dependencies
                # tensor_dependencies[0] corresponds to the prior version
                # of the model.
                # If it exists locally, should pull the remote delta because
                # this is the least costly path
                prior_model_layer = self.tensor_db.get_tensor_from_cache(tensor_dependencies[0])
                if prior_model_layer is not None:
                    uncompressed_delta = self.get_aggregated_tensor_from_aggregator(
                        tensor_dependencies[1]
                    )
                    new_model_tk, nparray = self.tensor_codec.apply_delta(
                        tensor_dependencies[1],
                        uncompressed_delta,
                        prior_model_layer,
                        creates_model=True,
                    )
                    self.tensor_db.cache_tensor({new_model_tk: nparray})
                else:
                    logger.info(
                        "Could not find previous model layer.Fetching latest layer from aggregator"
                    )
                    # The original model tensor should be fetched from aggregator
                    nparray = self.get_aggregated_tensor_from_aggregator(
                        tensor_key, require_lossless=True
                    )
            elif "model" in tags:
                # Pulling the model for the first time
                nparray = self.get_aggregated_tensor_from_aggregator(
                    tensor_key, require_lossless=True
                )
            else:
                # we should try fetching the tensor from aggregator
                tensor_name, origin, round_number, report, tags = tensor_key
                tags = (self.collaborator_name,) + tags
                tensor_key = (tensor_name, origin, round_number, report, tags)
                logger.info(
                    "Could not find previous model layer."
                    f"Fetching latest layer from aggregator {tensor_key}"
                )
                nparray = self.get_aggregated_tensor_from_aggregator(
                    tensor_key, require_lossless=True
                )
        else:
            logger.debug("Found tensor %s in local TensorDB", tensor_key)

        return nparray

    def get_aggregated_tensor_from_aggregator(self, tensor_key, require_lossless=False):
        """
        Return the decompressed tensor associated with the requested tensor key.

        If the key requests a compressed tensor (in the tag), the tensor will
        be decompressed before returning.
        If the key specifies an uncompressed tensor (or just omits a compressed
        tag), the decompression operation will be skipped.

        Args:
            tensor_key (namedtuple): The requested tensor.
            require_lossless (bool): Should compression of the tensor be
                allowed in flight? For the initial model, it may affect
                convergence to apply lossy compression. And metrics shouldn't
                be compressed either.

        Returns:
            nparray : The decompressed tensor associated with the requested
                tensor key.
        """
        tensor_name, origin, round_number, report, tags = tensor_key

        logger.debug("Requesting aggregated tensor %s", tensor_key)
        tensor = self.client.get_aggregated_tensor(
            tensor_name,
            round_number,
            report,
            tags,
            require_lossless,
        )

        # this translates to a numpy array and includes decompression, as
        # necessary
        nparray = self.named_tensor_to_nparray(tensor)

        # cache this tensor
        self.tensor_db.cache_tensor({tensor_key: nparray})

        return nparray

    def send_task_results(self, tensor_dict, round_number, task_name) -> dict:
        """Send task results to the aggregator.

        Args:
            tensor_dict (dict): Tensor dictionary.
            round_number (int):  Actual round number.
            task_name (string): Task name.

        Returns:
            A dictionary of reportable metrics of the current collaborator for the task.
        """
        named_tensors = [self.nparray_to_named_tensor(k, v) for k, v in tensor_dict.items()]

        # for general tasks, there may be no notion of data size to send.
        # But that raises the question how to properly aggregate results.

        data_size = -1

        if "train" in task_name:
            data_size = self.task_runner.get_train_data_size()

        if "valid" in task_name:
            data_size = self.task_runner.get_valid_data_size()

        logger.debug("%s data size = %s", task_name, data_size)

        metrics = {}
        for tensor in tensor_dict:
            tensor_name, origin, fl_round, report, tags = tensor

            if report:
                # Reportable metric must be a scalar
                value = float(tensor_dict[tensor])
                metrics.update({f"{self.collaborator_name}/{task_name}/{tensor_name}": value})

        self.client.send_local_task_results(
            round_number,
            task_name,
            data_size,
            named_tensors,
        )

        return metrics

    def nparray_to_named_tensor(self, tensor_key, nparray):
        """Construct the NamedTensor Protobuf.

        Includes logic to create delta, compress tensors with the TensorCodec,
        etc.

        Args:
            tensor_key (namedtuple): Tensorkey that will be resolved locally or
                remotely. May be the product of other tensors.
            nparray: The decompressed tensor associated with the requested
                tensor key.

        Returns:
            named_tensor (protobuf) : The tensor constructed from the nparray.
        """
        # if we have an aggregated tensor, we can make a delta
        tensor_name, origin, round_number, report, tags = tensor_key
        if "trained" in tags and self.use_delta_updates:
            # Should get the pretrained model to create the delta. If training
            # has happened,
            # Model should already be stored in the TensorDB
            model_nparray = self.tensor_db.get_tensor_from_cache(
                TensorKey(tensor_name, origin, round_number, report, ("model",))
            )

            # The original model will not be present for the optimizer on the
            # first round.
            if model_nparray is not None:
                delta_tensor_key, delta_nparray = self.tensor_codec.generate_delta(
                    tensor_key, nparray, model_nparray
                )
                delta_comp_tensor_key, delta_comp_nparray, metadata = self.tensor_codec.compress(
                    delta_tensor_key, delta_nparray
                )

                named_tensor = utils.construct_named_tensor(
                    delta_comp_tensor_key,
                    delta_comp_nparray,
                    metadata,
                    lossless=False,
                )
                return named_tensor

        # Assume every other tensor requires lossless compression
        compressed_tensor_key, compressed_nparray, metadata = self.tensor_codec.compress(
            tensor_key, nparray, require_lossless=True
        )
        named_tensor = utils.construct_named_tensor(
            compressed_tensor_key, compressed_nparray, metadata, lossless=True
        )

        return named_tensor

    def named_tensor_to_nparray(self, named_tensor):
        """Convert named tensor to a numpy array.

        Args:
            named_tensor (protobuf): The tensor to convert to nparray.

        Returns:
            decompressed_nparray (nparray): The nparray converted.
        """
        # do the stuff we do now for decompression and frombuffer and stuff
        # This should probably be moved back to protoutils
        raw_bytes = named_tensor.data_bytes
        metadata = [
            {
                "int_to_float": proto.int_to_float,
                "int_list": proto.int_list,
                "bool_list": proto.bool_list,
            }
            for proto in named_tensor.transformer_metadata
        ]
        # The tensor has already been transferred to collaborator, so
        # the newly constructed tensor should have the collaborator origin
        tensor_key = TensorKey(
            named_tensor.name,
            self.collaborator_name,
            named_tensor.round_number,
            named_tensor.report,
            tuple(named_tensor.tags),
        )
        *_, tags = tensor_key
        if "compressed" in tags:
            decompressed_tensor_key, decompressed_nparray = self.tensor_codec.decompress(
                tensor_key,
                data=raw_bytes,
                transformer_metadata=metadata,
                require_lossless=True,
            )
        elif "lossy_compressed" in tags:
            decompressed_tensor_key, decompressed_nparray = self.tensor_codec.decompress(
                tensor_key, data=raw_bytes, transformer_metadata=metadata
            )
        else:
            # There could be a case where the compression pipeline is bypassed
            # entirely
            logger.warning("Bypassing tensor codec...")
            decompressed_tensor_key = tensor_key
            decompressed_nparray = raw_bytes

        self.tensor_db.cache_tensor({decompressed_tensor_key: decompressed_nparray})

        return decompressed_nparray

    def _apply_masks(
        self,
        tensor_dict,
    ):
        """
        Calculate masked input vectors for secure aggregation.

        This function fetches private and shared masks from the tensor database if
        they are not provided, and applies these masks to the input tensors.

        Args:
            tensor_dict (dict): A dictionary of tensors to be masked.
        """
        import numpy as np

        # Fetch private mask from tensor db if not already fetched.
        if not self._private_mask:
            self._private_mask = self.tensor_db.get_tensor_from_cache(
                TensorKey("private_mask", self.collaborator_name, -1, False, ("secagg",))
            )[0]
        # Fetch shared mask from tensor db if not already fetched.
        if not self._shared_mask:
            self._shared_mask = self.tensor_db.get_tensor_from_cache(
                TensorKey("shared_mask", self.collaborator_name, -1, False, ("secagg",))
            )[0]

        for tensor_key in tensor_dict:
            _, _, _, _, tags = tensor_key
            if "metric" in tags:
                continue
            masked_metric = np.add(self._private_mask, tensor_dict[tensor_key])
            tensor_dict[tensor_key] = np.add(masked_metric, self._shared_mask)



#!/usr/bin/env python
# coding: utf-8

# Copyright (C) 2022-2024 TU Darmstadt
# SPDX-License-Identifier: Apache-2.0

# -----------------------------------------------------------
# Primary author: Phillip Rieger <phillip.rieger@trust.tu-darmstadt.de>
# Co-authored-by: Torsten Krauss <torsten.krauss@uni-wuerzburg.de>
# ------------------------------------------------------------

import argparse
import random
import time
import warnings

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import TensorDataset
import torch.optim as optim
from torchvision import transforms, datasets
from sklearn.cluster import AgglomerativeClustering, DBSCAN

from CrowdGuardClientValidation import CrowdGuardClientValidation
from openfl.experimental.workflow.interface import Aggregator, Collaborator, FLSpec
from openfl.experimental.workflow.placement import aggregator, collaborator
from openfl.experimental.workflow.runtime import LocalRuntime
from urllib.request import urlretrieve

warnings.filterwarnings("ignore")

BATCH_SIZE_TRAIN = 32
BATCH_SIZE_TEST = 1000
LEARNING_RATE = 0.00075
MOMENTUM = 0.9
LOG_INTERVAL = 10
TOTAL_CLIENT_NUMBER = 4
PMR = 0.25
NUMBER_OF_MALICIOUS_CLIENTS = max(1, int(TOTAL_CLIENT_NUMBER * PMR)) if PMR > 0 else 0
NUMBER_OF_BENIGN_CLIENTS = TOTAL_CLIENT_NUMBER - NUMBER_OF_MALICIOUS_CLIENTS
PRETRAINED_MODEL_FILE = 'pretrained_cifar.pt'

# set the random seed for repeatable results
RANDOM_SEED = 10

VOTE_FOR_BENIGN = 1
VOTE_FOR_POISONED = 0
STD_DEV = torch.from_numpy(np.array([0.2023, 0.1994, 0.2010]))
MEAN = torch.from_numpy(np.array([0.4914, 0.4822, 0.4465]))


def download_pretrained_model():
    urlretrieve('https://huggingface.co/prieger/cifar10/resolve/main/pretrained_cifar.pt?'
                'download=true', PRETRAINED_MODEL_FILE)


def trigger_single_image(image):
    """
    Adds a red square with a height/width of 6 pixels into
    the upper left corner of the given image.
    @param image tensor, containing the normalized pixel values of the image.
    The image will be modified in-place.
    @return given image
    """
    color = (torch.Tensor((1, 0, 0)) - MEAN) / STD_DEV
    image[:, 0:6, 0:6] = color.repeat((6, 6, 1)).permute(2, 1, 0)
    return image


def poison_data(samples_to_poison, labels_to_poison, pdr=0.5):
    """
    poisons a given local dataset, consisting of samples and labels, s.t.,
    the given ratio of this image consists of samples for the backdoor behavior
    :param samples_to_poison tensor containing all samples of the local dataset
    :param labels_to_poison tensor containing all labels
    :param pdr poisoned data rate
    :return poisoned local dataset (samples, labels)
    """
    if pdr == 0:
        return samples_to_poison, labels_to_poison

    assert 0 < pdr <= 1.0
    samples_to_poison = samples_to_poison.clone()
    labels_to_poison = labels_to_poison.clone()

    dataset_size = samples_to_poison.shape[0]
    num_samples_to_poison = int(dataset_size * pdr)
    if num_samples_to_poison == 0:
        # corner case for tiny pdrs
        assert pdr > 0  # Already checked above
        assert dataset_size > 1
        num_samples_to_poison += 1

    indices = np.random.choice(dataset_size, size=num_samples_to_poison, replace=False)
    for image_index in indices:
        image = trigger_single_image(samples_to_poison[image_index])
        samples_to_poison[image_index] = image
    labels_to_poison[indices] = 2
    return samples_to_poison, labels_to_poison.long()


class SequentialWithInternalStatePrediction(nn.Sequential):
    """
    Adapted version of Sequential that implements the function predict_internal_states
    """

    def predict_internal_states(self, x):
        """
        applies the submodules on the input. Compared to forward, this function also returns
        all intermediate outputs
        """
        result = []
        for module in self:
            x = module(x)
            # We can define our layer as we want. We selected Convolutional and
            # Linear Modules as layers here.
            # Differs for every model architecture.
            # Can be defined by the defender.
            if isinstance(module, nn.Conv2d) or isinstance(module, nn.Linear):
                result.append(x)
        return result, x


class Net(nn.Module):
    def __init__(self, num_classes=10):
        super(Net, self).__init__()
        self.features = SequentialWithInternalStatePrediction(
            nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
            nn.Conv2d(64, 192, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
        )
        self.classifier = SequentialWithInternalStatePrediction(
            nn.Dropout(),
            nn.Linear(256 * 2 * 2, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, num_classes),
        )

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), 256 * 2 * 2)
        x = self.classifier(x)
        return x

    def predict_internal_states(self, x):
        result, x = self.features.predict_internal_states(x)
        x = x.view(x.size(0), 256 * 2 * 2)
        result += self.classifier.predict_internal_states(x)[0]
        return result


def default_optimizer(model, optimizer_type=None, optimizer_like=None):
    """
    Return a new optimizer based on the optimizer_type or the optimizer template

    Args:
        model:   NN model architected from nn.module class
        optimizer_type: "SGD" or "Adam"
        optimizer_like: "torch.optim.SGD" or "torch.optim.Adam" optimizer
    """
    if optimizer_type == "SGD" or isinstance(optimizer_like, optim.SGD):
        return optim.SGD(model.parameters(), lr=LEARNING_RATE, momentum=MOMENTUM)
    elif optimizer_type == "Adam" or isinstance(optimizer_like, optim.Adam):
        return optim.Adam(model.parameters())


def test(network, test_loader, device, mode='Benign', move_to_cpu_afterward=True,
         test_train='Test'):
    network.eval()
    network.to(device)
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data = data.to(device)
            target = target.to(device)
            output = network(data)
            criterion = nn.CrossEntropyLoss()
            test_loss += criterion(output, target).item()
            pred = output.data.max(1, keepdim=True)[1]
            correct += pred.eq(target.data.view_as(pred)).sum()
    test_loss /= len(test_loader)
    accuracy = float(correct / len(test_loader.dataset))
    print(
        (
            f"{mode} {test_train} set: Avg. loss: {test_loss}, "
            f"Accuracy: {correct}/{len(test_loader.dataset)} ({100.0 * accuracy:5.03f}%)"
        )
    )
    if move_to_cpu_afterward:
        network.to("cpu")
    return accuracy


def FedAvg(models):  # NOQA: N802
    """
    Return a Federated average model based on Fedavg algorithm: H. B. Mcmahan,
    E. Moore, D. Ramage, S. Hampson, and B. A. Y.Arcas,
    “Communication-efficient learning of deep networks from decentralized data,” 2017.

    Args:
        models: Python list of locally trained models by each collaborator
    """
    new_model = models[0]
    if len(models) > 1:
        state_dicts = [model.state_dict() for model in models]
        state_dict = new_model.state_dict()
        for key in models[1].state_dict():
            state_dict[key] = torch.from_numpy(
                np.average([state[key].numpy() for state in state_dicts], axis=0))
        new_model.load_state_dict(state_dict)
    return new_model


def scale_update_of_model(to_scale, global_model, scaling_factor):
    """
    Scales the update of a local model (thus the difference between global and local model)
    :param to_scale: local model as state dict
    :pram global_model
    :param scaling factor
    :return scaled local model as state dict
    """
    print(f'Scale Model by {scaling_factor}')
    result = {}
    for name, data in to_scale.items():
        if not (name.endswith('.bias') or name.endswith('.weight')):
            result[name] = data
        else:
            update = data - global_model[name]
            scaled = scaling_factor * update
            result[name] = scaled + global_model[name]
    return result


def create_cluster_map_from_labels(expected_number_of_labels, clustering_labels):
    """
    Converts a list of labels into a dictionary where each label is the key and
    the values are lists/np arrays of the indices from the samples that received
    the respective label
    :param expected_number_of_labels number of samples whose labels are contained in
    clustering_labels
    :param clustering_labels list containing the labels of each sample
    :return dictionary of clusters
    """
    assert len(clustering_labels) == expected_number_of_labels

    clusters = {}
    for i, cluster in enumerate(clustering_labels):
        if cluster not in clusters:
            clusters[cluster] = []
        clusters[cluster].append(i)
    return {index: np.array(cluster) for index, cluster in clusters.items()}


def determine_biggest_cluster(clustering):
    """
    Given a clustering, given as dictionary of the form {cluster_id: [items in cluster]}, the
    function returns the id of the biggest cluster
    """
    biggest_cluster_id = None
    biggest_cluster_size = None
    for cluster_id, cluster in clustering.items():
        size_of_current_cluster = np.array(cluster).shape[0]
        if biggest_cluster_id is None or size_of_current_cluster > biggest_cluster_size:
            biggest_cluster_id = cluster_id
            biggest_cluster_size = size_of_current_cluster
    return biggest_cluster_id


class FederatedFlow(FLSpec):
    def __init__(self, model, optimizers, device="cpu", total_rounds=10, top_model_accuracy=0,
                 pmr=0.25, aggregation_algorithm='FedAVG', **kwargs, ):
        if aggregation_algorithm not in ['FedAVG', 'CrowdGuard']:
            raise Exception(f'Unsupported Aggregation Algorithm: {aggregation_algorithm}')
        super().__init__(**kwargs)
        self.aggregation_algorithm = aggregation_algorithm
        self.model = model
        self.global_model = Net()
        self.pmr = pmr
        self.start_time = None
        self.collaborators = None
        self.private = None
        self.optimizers = optimizers
        self.total_rounds = total_rounds
        self.top_model_accuracy = top_model_accuracy
        self.device = device
        self.round_num = 0  # starting round
        print(20 * "#")
        print(f"Round {self.round_num}...")
        print(20 * "#")

    @aggregator
    def start(self):
        self.start_time = time.time()
        print("Performing initialization for model")
        self.collaborators = self.runtime.collaborators
        self.private = 10
        self.next(
            self.train,
            foreach="collaborators",
            exclude=["private"],
        )

    @collaborator
    def train(self):
        self.collaborator_name = self.input
        print(20 * "#")
        print(f"Performing model training for collaborator {self.input} in round {self.round_num}")

        self.model.to(self.device)
        original_model = {n: d.clone() for n, d in self.model.state_dict().items()}
        test(self.model, self.train_loader, self.device, move_to_cpu_afterward=False,
             test_train='Train')
        test(self.model, self.test_loader, self.device, move_to_cpu_afterward=False)
        test(self.model, self.backdoor_test_loader, self.device, mode='Backdoor',
             move_to_cpu_afterward=False)
        self.optimizer = default_optimizer(self.model, optimizer_like=self.optimizers[self.input])

        self.model.train()
        train_losses = []
        for batch_idx, (data, target) in enumerate(self.train_loader):
            data = data.to(self.device)
            target = target.to(self.device)
            self.optimizer.zero_grad()
            output = self.model(data)
            criterion = nn.CrossEntropyLoss()
            loss = criterion(output, target).to(self.device)
            loss.backward()
            self.optimizer.step()
            if batch_idx % LOG_INTERVAL == 0:
                train_losses.append(loss.item())

        self.loss = np.mean(train_losses)
        self.training_completed = True

        test(self.model, self.train_loader, self.device, move_to_cpu_afterward=False,
             test_train='Train')
        test(self.model, self.test_loader, self.device, move_to_cpu_afterward=False)
        test(self.model, self.backdoor_test_loader, self.device, mode='Backdoor',
             move_to_cpu_afterward=False)
        if 'malicious' in self.input:
            weights = self.model.state_dict()
            scaled = scale_update_of_model(weights, original_model, 1 / self.pmr)
            self.model.load_state_dict(scaled)
        self.model.to("cpu")
        torch.cuda.empty_cache()
        if self.aggregation_algorithm == 'FedAVG':
            self.next(self.fed_avg_aggregation, exclude=["training_completed"])
        else:
            self.next(self.collect_models, exclude=["training_completed"])

    @aggregator
    def fed_avg_aggregation(self, inputs):
        self.all_models = {input.collaborator_name: input.model.cpu() for input in inputs}
        self.model = FedAvg([m.cpu() for m in self.all_models.values()])
        self.round_num += 1
        if self.round_num + 1 < self.total_rounds:
            self.next(self.train, foreach="collaborators")
        else:
            self.next(self.end)

    @aggregator
    def collect_models(self, inputs):
        # Following the CrowdGuard paper, this should be executed within SGX

        self.all_models = {i.collaborator_name: i.model.cpu() for i in inputs}
        self.next(self.local_validation, foreach="collaborators")

    @collaborator
    def local_validation(self):
        # Following the CrowdGuard paper, this should be executed within SGX
        print(
            f"Performing model validation for collaborator {self.input} in round {self.round_num}"
        )
        self.collaborator_name = self.input
        all_names = list(self.all_models.keys())
        all_models = [self.all_models[n] for n in all_names]
        own_client_index = all_names.index(self.collaborator_name)
        detected_suspicious_models = CrowdGuardClientValidation.validate_models(self.global_model,
                                                                                all_models,
                                                                                own_client_index,
                                                                                self.train_loader,
                                                                                self.device)
        detected_suspicious_models = sorted(detected_suspicious_models)
        print(
            f'Suspicious Models detected by {own_client_index}: {detected_suspicious_models}')

        votes_of_this_client = []
        for c in range(len(all_models)):
            if c == own_client_index:
                votes_of_this_client.append(VOTE_FOR_BENIGN)
            elif c in detected_suspicious_models:
                votes_of_this_client.append(VOTE_FOR_POISONED)
            else:
                votes_of_this_client.append(VOTE_FOR_BENIGN)
        self.votes_of_this_client = {}
        for name, vote in zip(all_names, votes_of_this_client):
            self.votes_of_this_client[name] = vote

        self.next(self.defend)

    @aggregator
    def defend(self, inputs):
        # Following the CrowdGuard paper, this should be executed within SGX

        all_names = list(self.all_models.keys())
        all_votes_by_name = {i.collaborator_name: i.votes_of_this_client for i in inputs}

        all_models = [self.all_models[name] for name in all_names]
        binary_votes = [[all_votes_by_name[own_name][val_name] for val_name in all_names] for
                        own_name in all_names]

        ac_e = AgglomerativeClustering(n_clusters=2, distance_threshold=None,
                                       compute_full_tree=True,
                                       metric="euclidean", memory=None, connectivity=None,
                                       linkage='single',
                                       compute_distances=True).fit(binary_votes)
        ac_e_labels: list = ac_e.labels_.tolist()
        agglomerative_result = create_cluster_map_from_labels(len(all_names), ac_e_labels)
        print(f'Agglomerative Clustering: {agglomerative_result}')
        agglomerative_negative_cluster = agglomerative_result[
            determine_biggest_cluster(agglomerative_result)]

        db_scan_input_idx_list = agglomerative_negative_cluster
        print(f'DBScan Input: {db_scan_input_idx_list}')
        db_scan_input_list = [binary_votes[vote_id] for vote_id in db_scan_input_idx_list]

        db = DBSCAN(eps=0.5, min_samples=1).fit(db_scan_input_list)
        dbscan_clusters = create_cluster_map_from_labels(len(agglomerative_negative_cluster),
                                                         db.labels_.tolist())
        biggest_dbscan_cluster = dbscan_clusters[determine_biggest_cluster(dbscan_clusters)]
        print(f'DBScan Clustering: {biggest_dbscan_cluster}')

        single_sample_of_biggest_cluster = biggest_dbscan_cluster[0]
        final_voting = db_scan_input_list[single_sample_of_biggest_cluster]
        negatives = [i for i, vote in enumerate(final_voting) if vote == VOTE_FOR_BENIGN]
        recognized_benign_models = [all_models[n] for n in negatives]

        print(f'Negatives: {negatives}')

        self.model = FedAvg([m.cpu() for m in recognized_benign_models])
        del inputs
        self.round_num += 1
        if self.round_num < self.total_rounds:
            print(f'Finished round {self.round_num}/{self.total_rounds}')
            self.next(self.train, foreach="collaborators")
        else:
            self.next(self.end)

    @aggregator
    def end(self):
        print(20 * "#")
        print("All rounds completed successfully")
        print(20 * "#")
        print("This is the end of the flow")
        print(20 * "#")


def seed_random_generators(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


if __name__ == '__main__':

    seed_random_generators(RANDOM_SEED)

    argparser = argparse.ArgumentParser(description=__doc__)
    argparser.add_argument(
        "--test_dataset_ratio",
        type=float,
        default=0.4,
        help="Indicate the what fraction of the sample will be used for testing",
    )
    argparser.add_argument(
        "--train_dataset_ratio",
        type=float,
        default=0.4,
        help="Indicate the what fraction of the sample will be used for training",
    )

    argparser.add_argument(
        "--log_dir",
        type=str,
        default="test_debug",
        help="Indicate where to save the privacy loss profile and log files during the training",
    )
    argparser.add_argument(
        "--comm_round",
        type=int,
        default=30,
        help="Indicate the communication round of FL",
    )
    argparser.add_argument(
        "--optimizer_type",
        type=str,
        default="SGD",
        help="Indicate optimizer to use for training",
    )

    args = argparser.parse_args()

    download_pretrained_model()

    # Setup participants
    aggregator_object = Aggregator()
    aggregator_object.private_attributes = {}
    collaborator_names = [f'benign_{i:02d}' for i in range(NUMBER_OF_BENIGN_CLIENTS)] + [
        f'malicious_{i:02d}' for i in range(NUMBER_OF_MALICIOUS_CLIENTS)]
    collaborators = [Collaborator(name=name) for name in collaborator_names]
    if torch.cuda.is_available():
        device = torch.device(
            "cuda:1"
        )  # This will enable Ray library to reserve available GPU(s) for the task
    else:
        device = torch.device("cpu")

    # Prepare local datasets
    transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize(MEAN, STD_DEV), ])
    cifar_train = datasets.CIFAR10(root="./data", train=True, download=True, transform=transform)
    cifar_train = list(cifar_train)
    cifar_test = datasets.CIFAR10(root="./data", train=False, download=True, transform=transform)
    cifar_test = list(cifar_test)
    X = torch.stack([x[0] for x in cifar_train] + [x[0] for x in cifar_test])
    Y = torch.LongTensor(
        np.stack(np.array([x[1] for x in cifar_train] + [x[1] for x in cifar_test])))

    # split the dataset
    seed_random_generators(RANDOM_SEED)
    shuffled_indices = np.arange(X.shape[0])
    np.random.shuffle(shuffled_indices)

    N_total_samples = len(cifar_test) + len(cifar_train)
    train_dataset_size = int(N_total_samples * args.train_dataset_ratio)
    test_dataset_size = int(N_total_samples * args.test_dataset_ratio)
    X = X[shuffled_indices]
    Y = Y[shuffled_indices]

    train_dataset_data = X[:train_dataset_size]
    train_dataset_targets = Y[:train_dataset_size]

    test_dataset_data = X[train_dataset_size:train_dataset_size + test_dataset_size]
    test_dataset_targets = Y[train_dataset_size:train_dataset_size + test_dataset_size]
    print(f"Dataset info (total {N_total_samples}): train - {test_dataset_targets.shape[0]}, "
          f"test - {test_dataset_targets.shape[0]}, ")

    # partition the dataset for clients

    for idx, collab in enumerate(collaborators):
        # construct the training and test and population dataset
        benign_training_x = train_dataset_data[idx::len(collaborators)]
        benign_training_y = train_dataset_targets[idx::len(collaborators)]

        if 'malicious' in collab.name:
            local_train_data, local_train_targets = poison_data(benign_training_x,
                                                                benign_training_y)
        else:
            local_train_data, local_train_targets = benign_training_x, benign_training_y

        local_test_data = test_dataset_data[idx::len(collaborators)]
        local_test_targets = test_dataset_targets[idx::len(collaborators)]

        poison_test_data, poison_test_targets = poison_data(local_test_data, local_test_targets,
                                                            pdr=1.0)

        collab.private_attributes = {
            "train_loader": torch.utils.data.DataLoader(
                TensorDataset(local_train_data, local_train_targets),
                batch_size=BATCH_SIZE_TRAIN, shuffle=True
            ),
            "test_loader": torch.utils.data.DataLoader(
                TensorDataset(local_test_data, local_test_targets),
                batch_size=BATCH_SIZE_TEST, shuffle=False
            ),
            "backdoor_test_loader": torch.utils.data.DataLoader(
                TensorDataset(poison_test_data, poison_test_targets),
                batch_size=BATCH_SIZE_TEST, shuffle=False
            ),
        }

    local_runtime = LocalRuntime(aggregator=aggregator_object, collaborators=collaborators)

    print(f"Local runtime collaborators = {local_runtime.collaborators}")

    # change to the internal flow loop
    model = Net()
    top_model_accuracy = 0
    optimizers = {
        collaborator.name: default_optimizer(model, optimizer_type=args.optimizer_type)
        for collaborator in collaborators
    }
    flflow = FederatedFlow(
        model,
        optimizers,
        device,
        args.comm_round,
        top_model_accuracy,
        NUMBER_OF_MALICIOUS_CLIENTS / TOTAL_CLIENT_NUMBER,
        'CrowdGuard'
    )
    flflow.runtime = local_runtime
    flflow.run()



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""AggregatorGRPCServer module."""

import logging
from random import random
from time import sleep

import grpc

from openfl.protocols import aggregator_pb2, aggregator_pb2_grpc, utils
from openfl.transport.grpc.common import create_grpc_server, create_header, synchronized

logger = logging.getLogger(__name__)


class AggregatorGRPCServer(aggregator_pb2_grpc.AggregatorServicer):
    """Aggregator gRPC Server.

    Attributes:
        aggregator (Aggregator): An instance of the Aggregator object that this server is serving.
        agg_port (int): Port to start gRPC server on.
        use_tls (bool): Whether to use TLS for the connection.
        require_client_auth (bool): Whether to enable client-side authentication, i.e. mTLS.
            Ignored if `use_tls=False`.
        root_certificate (str): The path to the root certificate for the TLS connection, ignored if
            `use_tls=False`.
        certificate (str): The path to the client's certificate for the TLS connection, ignored if
            `use_tls=False`.
        private_key (str): The path to the client's private key for the TLS connection, ignored if
            `use_tls=False`.
        root_certificate_refresher_cb (Callable): A callback function that receives no arguments and
            returns the current root certificate.
    """

    def __init__(
        self,
        aggregator,
        agg_port,
        use_tls=True,
        require_client_auth=True,
        root_certificate=None,
        certificate=None,
        private_key=None,
        root_certificate_refresher_cb=None,
        **kwargs,
    ):
        self.aggregator = aggregator
        self.uri = f"[::]:{agg_port}"
        self.use_tls = use_tls
        self.require_client_auth = require_client_auth
        self.root_certificate = root_certificate
        self.certificate = certificate
        self.private_key = private_key

        if hasattr(self.aggregator, "is_connector_available"):
            self.use_connector = self.aggregator.is_connector_available()
        else:
            self.use_connector = False

        if self.use_connector:
            self.interop_client = (
                self.aggregator.get_interop_client()
            )  # Initialize the interoperability client
        else:
            self.interop_client = None

        self.root_certificate_refresher_cb = root_certificate_refresher_cb

    def validate_collaborator(self, request, context):
        """Validate the collaborator.

        This method checks that the collaborator who sent the request is
        authorized to do so.

        Args:
            request (aggregator_pb2.MessageHeader): The request from the
                collaborator.
            context (grpc.ServicerContext): The context of the request.

        Raises:
            grpc.RpcError: If the collaborator or collaborator certificate is
                not authorized.
        """
        if self.use_tls:
            collaborator_common_name = request.header.sender
            if self.require_client_auth:
                common_name = context.auth_context()["x509_common_name"][0].decode("utf-8")
            else:
                common_name = collaborator_common_name

            if not self.aggregator.valid_collaborator_cn_and_id(
                common_name, collaborator_common_name
            ):
                # Random delay in authentication failures
                sleep(5 * random())  # nosec
                context.abort(
                    grpc.StatusCode.UNAUTHENTICATED,
                    f"Invalid collaborator. CN: |{common_name}| "
                    f"collaborator_common_name: |{collaborator_common_name}|",
                )

    def check_request(self, request):
        """Validate request header matches expected values.

        This method checks that the request is valid and was sent by an
            authorized collaborator.

        Args:
            request (aggregator_pb2.MessageHeader): Request sent from a
                collaborator that requires validation.

        Raises:
            ValueError: If the request is not valid.
        """
        assert request.header.sender in self.aggregator.authorized_cols, (
            f"Sender in request header is not authorized. "
            f"Expected: one of {self.aggregator.authorized_cols}, Actual: {request.header.sender}"
        )

        assert request.header.receiver == self.aggregator.uuid, (
            f"Receiver in request header does not match aggregator UUID. "
            f"Expected: {self.aggregator.uuid}, Actual: {request.header.receiver}"
        )

        assert request.header.federation_uuid == self.aggregator.federation_uuid, (
            f"Federation UUID in request header does not match. "
            f"Expected: {self.aggregator.federation_uuid}, Actual: {request.header.federation_uuid}"
        )

        assert (
            request.header.single_col_cert_common_name
            == self.aggregator.single_col_cert_common_name
        ), (
            f"Single collaborator certificate common name in request header does not match. "
            f"Expected: {self.aggregator.single_col_cert_common_name}, Actual: {request.header.single_col_cert_common_name}"  # noqa: E501
        )

    def Ping(self, request, context):  # NOQA:N802
        """Ping endpoint of the Aggregator server.

        This method handles a ping request from a collaborator.

        Args:
            request (aggregator_pb2.PingRequest): The ping request from the
                collaborator.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            aggregator_pb2.PingResponse: The response to the ping request.
        """
        self.validate_collaborator(request, context)
        self.check_request(request)
        header = create_header(
            sender=self.aggregator.uuid,
            receiver=request.header.sender,
            federation_uuid=self.aggregator.federation_uuid,
            single_col_cert_common_name=self.aggregator.single_col_cert_common_name,
        )

        return aggregator_pb2.PingResponse(header=header)

    def GetTasks(self, request, context):  # NOQA:N802
        """Request a job from aggregator.

        This method handles a request from a collaborator for a job.

        Args:
            request (aggregator_pb2.GetTasksRequest): The request from the
                collaborator.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            aggregator_pb2.GetTasksResponse: The response to the request.
        """
        self.validate_collaborator(request, context)
        self.check_request(request)
        collaborator_name = request.header.sender
        tasks, round_number, sleep_time, time_to_quit = self.aggregator.get_tasks(
            request.header.sender
        )
        if tasks:
            if isinstance(tasks[0], str):
                # backward compatibility
                tasks_proto = [
                    aggregator_pb2.Task(
                        name=task,
                    )
                    for task in tasks
                ]
            else:
                tasks_proto = [
                    aggregator_pb2.Task(
                        name=task.name,
                        function_name=task.function_name,
                        task_type=task.task_type,
                        apply_local=task.apply_local,
                    )
                    for task in tasks
                ]
        else:
            tasks_proto = []

        header = create_header(
            sender=self.aggregator.uuid,
            receiver=collaborator_name,
            federation_uuid=self.aggregator.federation_uuid,
            single_col_cert_common_name=self.aggregator.single_col_cert_common_name,
        )

        return aggregator_pb2.GetTasksResponse(
            header=header,
            round_number=round_number,
            tasks=tasks_proto,
            sleep_time=sleep_time,
            quit=time_to_quit,
        )

    def GetAggregatedTensor(self, request, context):  # NOQA:N802
        """Request a job from aggregator.

        This method handles a request from a collaborator for an aggregated
        tensor.

        Args:
            request (aggregator_pb2.GetAggregatedTensorRequest): The request
                from the collaborator.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            aggregator_pb2.GetAggregatedTensorResponse: The response to the
                request.
        """
        if self.use_connector:
            context.abort(
                grpc.StatusCode.UNIMPLEMENTED,
                "This method is not available in framework interopability mode.",
            )

        self.validate_collaborator(request, context)
        self.check_request(request)

        named_tensor = self.aggregator.get_aggregated_tensor(
            request.tensor_name,
            request.round_number,
            request.report,
            tuple(request.tags),
            request.require_lossless,
        )

        header = create_header(
            sender=self.aggregator.uuid,
            receiver=request.header.sender,
            federation_uuid=self.aggregator.federation_uuid,
            single_col_cert_common_name=self.aggregator.single_col_cert_common_name,
        )

        return aggregator_pb2.GetAggregatedTensorResponse(
            header=header,
            round_number=request.round_number,
            tensor=named_tensor,
        )

    @synchronized
    def SendLocalTaskResults(self, request, context):  # NOQA:N802
        """Request a model download from aggregator.

        This method handles a request from a collaborator to send the results
        of a local task.

        Args:
            request (aggregator_pb2.SendLocalTaskResultsRequest): The request
                from the collaborator.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            aggregator_pb2.SendLocalTaskResultsResponse: The response to the
                request.
        """
        try:
            proto = aggregator_pb2.TaskResults()
            proto = utils.datastream_to_proto(proto, request)
        except RuntimeError:
            raise RuntimeError(
                "Empty stream message, reestablishing connection from client to resume training..."
            )

        self.validate_collaborator(proto, context)
        # all messages get sanity checked
        self.check_request(proto)

        collaborator_name = proto.header.sender
        task_name = proto.task_name
        round_number = proto.round_number
        data_size = proto.data_size
        named_tensors = proto.tensors
        self.aggregator.send_local_task_results(
            collaborator_name, round_number, task_name, data_size, named_tensors
        )
        # turn data stream into local model update
        header = create_header(
            sender=self.aggregator.uuid,
            receiver=collaborator_name,
            federation_uuid=self.aggregator.federation_uuid,
            single_col_cert_common_name=self.aggregator.single_col_cert_common_name,
        )

        return aggregator_pb2.SendLocalTaskResultsResponse(header=header)

    def InteropRelay(self, request, context):
        """
        Args:
            request (aggregator_pb2.InteropRelay): The request
                from the collaborator.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            aggregator_pb2.InteropRelay: The response to the
            request.
        """
        if not self.use_connector:
            context.abort(
                grpc.StatusCode.UNIMPLEMENTED,
                "InteropRelay is only available in federated interopability mode.",
            )

        self.validate_collaborator(request, context)
        self.check_request(request)
        collaborator_name = request.header.sender

        header = create_header(
            sender=self.aggregator.uuid,
            receiver=collaborator_name,
            federation_uuid=self.aggregator.federation_uuid,
            single_col_cert_common_name=self.aggregator.single_col_cert_common_name,
        )

        # Forward the incoming OpenFL message to the local gRPC client
        return self.interop_client.send_receive(request, header=header)

    def serve(self):
        """Starts the aggregator gRPC server."""

        if self.use_connector:
            self.aggregator.start_connector()

        server = create_grpc_server(
            self.uri,
            self.use_tls,
            self.private_key,
            self.certificate,
            self.root_certificate,
            self.require_client_auth,
            self.root_certificate_refresher_cb,
        )
        aggregator_pb2_grpc.add_AggregatorServicer_to_server(self, server)

        logger.info("Starting Aggregator gRPC Server")
        server.start()

        while not self.aggregator.all_quit_jobs_sent():
            sleep(5)

        if self.use_connector:
            self.aggregator.stop_connector()

        server.stop(0)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Aggregator module."""

import os
import sys
from logging import getLogger
from pathlib import Path

import yaml
from click import Path as ClickPath
from click import confirm, echo, group, option, pass_context, style

from openfl.cryptography.ca import sign_certificate
from openfl.cryptography.io import get_csr_hash, read_crt, read_csr, read_key, write_crt, write_key
from openfl.cryptography.participant import generate_csr
from openfl.experimental.workflow.federated.plan import Plan
from openfl.experimental.workflow.interface.cli.cli_helper import CERT_DIR
from openfl.utilities import click_types
from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.utils import getfqdn_env

logger = getLogger(__name__)


@group()
@pass_context
def aggregator(context):
    """Manage Federated Learning Aggregator."""
    context.obj["group"] = "aggregator"


@aggregator.command(name="start")
@option(
    "-p",
    "--plan",
    required=False,
    help="Federated learning plan [plan/plan.yaml]",
    default="plan/plan.yaml",
    type=ClickPath(exists=True),
)
@option(
    "-c",
    "--authorized_cols",
    required=False,
    help="Authorized collaborator list [plan/cols.yaml]",
    default="plan/cols.yaml",
    type=ClickPath(exists=True),
)
@option(
    "-s",
    "--secure",
    required=False,
    help="Enable Intel SGX Enclave",
    is_flag=True,
    default=False,
)
def start_(plan, authorized_cols, secure):
    """Start the aggregator service."""

    if is_directory_traversal(plan):
        echo("Federated learning plan path is out of the openfl workspace scope.")
        sys.exit(1)
    if is_directory_traversal(authorized_cols):
        echo("Authorized collaborator list file path is out of the openfl workspace scope.")
        sys.exit(1)

    plan = Plan.parse(
        plan_config_path=Path(plan).absolute(),
        cols_config_path=Path(authorized_cols).absolute(),
    )

    if not os.path.exists("plan/data.yaml"):
        logger.warning(
            "Aggregator private attributes are set to None as plan/data.yaml not found"
            + " in workspace."
        )
    else:
        with open("plan/data.yaml", "r") as f:
            data = yaml.safe_load(f)
            if data.get("aggregator", None) is None:
                logger.warning(
                    "Aggregator private attributes are set to None as no aggregator"
                    + " attributes found in plan/data.yaml."
                )

    logger.info("🧿 Starting the Aggregator Service.")

    agg_server = plan.get_server()
    agg_server.is_server_started = False
    agg_server.run_server()


@aggregator.command(name="generate-cert-request")
@option(
    "--fqdn",
    required=False,
    type=click_types.FQDN,
    help=f"The fully qualified domain name of aggregator node [{getfqdn_env()}]",
    default=getfqdn_env(),
)
def _generate_cert_request(fqdn):
    generate_cert_request(fqdn)


def generate_cert_request(fqdn):
    """Create aggregator certificate key pair."""

    if fqdn is None:
        fqdn = getfqdn_env()

    common_name = f"{fqdn}".lower()
    subject_alternative_name = f"DNS:{common_name}"
    file_name = f"agg_{common_name}"

    echo(
        f"Creating AGGREGATOR certificate key pair with following settings: "
        f"CN={style(common_name, fg='red')},"
        f" SAN={style(subject_alternative_name, fg='red')}"
    )

    server_private_key, server_csr = generate_csr(common_name, server=True)

    (CERT_DIR / "server").mkdir(parents=True, exist_ok=True)

    echo("  Writing AGGREGATOR certificate key pair to: " + style(f"{CERT_DIR}/server", fg="green"))

    # Print csr hash before writing csr to disk
    csr_hash = get_csr_hash(server_csr)
    echo("The CSR Hash " + style(f"{csr_hash}", fg="red"))

    # Write aggregator csr and key to disk
    write_crt(server_csr, CERT_DIR / "server" / f"{file_name}.csr")
    write_key(server_private_key, CERT_DIR / "server" / f"{file_name}.key")


@aggregator.command(name="certify")
@option(
    "-n",
    "--fqdn",
    type=click_types.FQDN,
    help=f"The fully qualified domain name of aggregator node [{getfqdn_env()}]",
    default=getfqdn_env(),
)
@option("-s", "--silent", help="Do not prompt", is_flag=True)
def _certify(fqdn, silent):
    certify(fqdn, silent)


def certify(fqdn, silent):
    """Sign/certify the aggregator certificate key pair."""

    if fqdn is None:
        fqdn = getfqdn_env()

    common_name = f"{fqdn}".lower()
    file_name = f"agg_{common_name}"
    cert_name = f"server/{file_name}"
    signing_key_path = "ca/signing-ca/private/signing-ca.key"
    signing_crt_path = "ca/signing-ca.crt"

    # Load CSR
    csr_path_absolute_path = Path(CERT_DIR / f"{cert_name}.csr").absolute()
    if not csr_path_absolute_path.exists():
        echo(
            style("Aggregator certificate signing request not found.", fg="red")
            + " Please run `fx aggregator generate-cert-request`"
            " to generate the certificate request."
        )

    csr, csr_hash = read_csr(csr_path_absolute_path)

    # Load private signing key
    private_sign_key_absolute_path = Path(CERT_DIR / signing_key_path).absolute()
    if not private_sign_key_absolute_path.exists():
        echo(
            style("Signing key not found.", fg="red") + " Please run `fx workspace certify`"
            " to initialize the local certificate authority."
        )

    signing_key = read_key(private_sign_key_absolute_path)

    # Load signing cert
    signing_crt_absolute_path = Path(CERT_DIR / signing_crt_path).absolute()
    if not signing_crt_absolute_path.exists():
        echo(
            style("Signing certificate not found.", fg="red") + " Please run `fx workspace certify`"
            " to initialize the local certificate authority."
        )

    signing_crt = read_crt(signing_crt_absolute_path)

    echo(
        "The CSR Hash for file "
        + style(f"{cert_name}.csr", fg="green")
        + " = "
        + style(f"{csr_hash}", fg="red")
    )

    crt_path_absolute_path = Path(CERT_DIR / f"{cert_name}.crt").absolute()

    if silent:
        echo(" Warning: manual check of certificate hashes is bypassed in silent mode.")
        echo(" Signing AGGREGATOR certificate")
        signed_agg_cert = sign_certificate(csr, signing_key, signing_crt.subject)
        write_crt(signed_agg_cert, crt_path_absolute_path)

    else:
        echo("Make sure the two hashes above are the same.")
        if confirm("Do you want to sign this certificate?"):
            echo(" Signing AGGREGATOR certificate")
            signed_agg_cert = sign_certificate(csr, signing_key, signing_crt.subject)
            write_crt(signed_agg_cert, crt_path_absolute_path)

        else:
            echo(
                style("Not signing certificate.", fg="red")
                + " Please check with this AGGREGATOR to get the correct"
                " certificate for this federation."
            )



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Plan module."""

import sys
from logging import getLogger
from pathlib import Path

from click import Path as ClickPath
from click import echo, group, option, pass_context

from openfl.experimental.workflow.federated import Plan
from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.utils import getfqdn_env

logger = getLogger(__name__)


@group()
@pass_context
def plan(context):
    """Manage Federated Learning Plans."""
    context.obj["group"] = "plan"


@plan.command()
@pass_context
@option(
    "-p",
    "--plan_config",
    required=False,
    help="Federated learning plan [plan/plan.yaml]",
    default="plan/plan.yaml",
    type=ClickPath(exists=True),
)
@option(
    "-c",
    "--cols_config",
    required=False,
    help="Authorized collaborator list [plan/cols.yaml]",
    default="plan/cols.yaml",
    type=ClickPath(exists=True),
)
@option(
    "-d",
    "--data_config",
    required=False,
    help="The data set/shard configuration file [plan/data.yaml]",
    default="plan/data.yaml",
)
@option(
    "-a",
    "--aggregator_address",
    required=False,
    help="The FQDN of the federation aggregator",
)
def initialize(context, plan_config, cols_config, data_config, aggregator_address):
    """
    Initialize Data Science plan.

    Create a protocol buffer file of the initial model weights for  the
    federation.
    """

    for p in [plan_config, cols_config, data_config]:
        if is_directory_traversal(p):
            echo(f"{p} is out of the openfl workspace scope.")
            sys.exit(1)

    plan_config = Path(plan_config).absolute()
    cols_config = Path(cols_config).absolute()
    data_config = Path(data_config).absolute()

    plan = Plan.parse(
        plan_config_path=plan_config,
        cols_config_path=cols_config,
        data_config_path=data_config,
    )

    plan_origin = Plan.parse(plan_config, resolve=False).config

    if plan_origin["network"]["settings"]["agg_addr"] == "auto" or aggregator_address:
        plan_origin["network"]["settings"]["agg_addr"] = aggregator_address or getfqdn_env()

        logger.warn(
            f"Patching Aggregator Addr in Plan 🠆 {plan_origin['network']['settings']['agg_addr']}"
        )

        Plan.dump(plan_config, plan_origin)

    plan.config = plan_origin

    # Record that plan with this hash has been initialized
    if "plans" not in context.obj:
        context.obj["plans"] = []
    context.obj["plans"].append(f"{plan_config.stem}_{plan.hash[:8]}")
    logger.info(f"{context.obj['plans']}")


def freeze_plan(plan_config):
    """Dump the plan to YAML file."""

    plan = Plan()
    plan.config = Plan.parse(Path(plan_config), resolve=False).config

    Plan.dump(Path(plan_config), plan.config, freeze=True)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Plan module."""

import sys
from logging import getLogger
from os import makedirs
from os.path import isfile
from pathlib import Path
from shutil import copyfile, rmtree

from click import Path as ClickPath
from click import echo, group, option, pass_context
from yaml import FullLoader, dump, load

from openfl.federated import Plan
from openfl.interface.cli_helper import get_workspace_parameter
from openfl.protocols import utils
from openfl.utilities.click_types import InputSpec
from openfl.utilities.dataloading import get_dataloader
from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.split import split_tensor_dict_for_holdouts
from openfl.utilities.utils import getfqdn_env

logger = getLogger(__name__)


@group()
@pass_context
def plan(context):
    """Manage Federated Learning Plans."""
    context.obj["group"] = "plan"


@plan.command()
@pass_context
@option(
    "-p",
    "--plan_config",
    required=False,
    help="Path to an FL plan.",
    default="plan/plan.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "-c",
    "--cols_config",
    required=False,
    help="Path to an authorized collaborator list.",
    default="plan/cols.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "-d",
    "--data_config",
    required=False,
    help="The dataset shard configuration file.",
    default="plan/data.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "-a",
    "--aggregator_address",
    required=False,
    help="The FQDN of the federation aggregator",
)
@option(
    "-f",
    "--input_shape",
    cls=InputSpec,
    required=False,
    help="""
    The input spec of the model.

    May be provided as a list for single input head: ``--input-shape [3,32,32]``,

    or as a dictionary for multihead models (must be passed in quotes):

    ``--input-shape "{'input_0': [1, 240, 240, 4],'input_1': [1, 240, 240, 1]}"``.
    """,
)
@option(
    "-g",
    "--gandlf_config",
    required=False,
    help="GaNDLF Configuration File Path",
)
@option(
    "-i",
    "--init_model_path",
    required=False,
    help="Path to initial model protobuf file.",
    type=ClickPath(exists=True),
)
def initialize(
    context,
    plan_config,
    cols_config,
    data_config,
    aggregator_address,
    input_shape,
    gandlf_config,
    init_model_path,
):
    """
    Initializes a Data Science plan and generates a protobuf file of the
    initial model weights for the federation.
    """

    for p in [plan_config, cols_config, data_config]:
        if is_directory_traversal(p):
            echo(f"{p} is out of the openfl workspace scope.")
            sys.exit(1)

    plan_config = Path(plan_config).absolute()
    cols_config = Path(cols_config).absolute()
    data_config = Path(data_config).absolute()
    if gandlf_config is not None:
        gandlf_config = Path(gandlf_config).absolute()

    plan = Plan.parse(
        plan_config_path=plan_config,
        cols_config_path=cols_config,
        data_config_path=data_config,
        gandlf_config_path=gandlf_config,
    )

    if "connector" in plan.config:
        logger.info("OpenFL Connector enabled: %s", plan.config["connector"])
        # Only need to initialize task runner to install apps/packages
        # that were not installable via requirements.txt
        plan.get_task_runner(data_loader=None)
    else:
        init_state_path = plan.config["aggregator"]["settings"]["init_state_path"]
        # This is needed to bypass data being locally available
        if input_shape is not None:
            logger.info(
                "Attempting to generate initial model weights with custom input shape "
                f"{input_shape}"
            )

        # Initialize tensor dictionary
        init_tensor_dict, task_runner, round_number = _initialize_tensor_dict(
            plan, input_shape, init_model_path
        )

        tensor_dict, holdout_params = split_tensor_dict_for_holdouts(
            init_tensor_dict,
            **task_runner.tensor_dict_split_fn_kwargs,
        )

        logger.warning(
            f"Following parameters omitted from global initial model, "
            f"local initialization will determine"
            f" values: {list(holdout_params.keys())}"
        )

        # Save the model state
        try:
            logger.info(f"Saving model state to {init_state_path}")
            plan.save_model_to_state_file(
                tensor_dict=tensor_dict, round_number=round_number, output_path=init_state_path
            )
        except Exception as e:
            logger.error(f"Failed to save model state: {e}")
            raise

    plan_origin = Plan.parse(
        plan_config_path=plan_config,
        gandlf_config_path=gandlf_config,
        resolve=False,
    )

    if plan_origin.config["network"]["settings"]["agg_addr"] == "auto" or aggregator_address:
        plan_origin.config["network"]["settings"]["agg_addr"] = aggregator_address or getfqdn_env()

        logger.warning(
            f"Patching Aggregator Addr in Plan"
            f" 🠆 {plan_origin.config['network']['settings']['agg_addr']}"
        )

        Plan.dump(plan_config, plan_origin.config)

    if gandlf_config is not None:
        Plan.dump(plan_config, plan_origin.config)

    # Record that plan with this hash has been initialized
    if "plans" not in context.obj:
        context.obj["plans"] = []
    context.obj["plans"].append(f"{plan_config.stem}_{plan_origin.hash[:8]}")
    logger.info(f"{context.obj['plans']}")


def _initialize_tensor_dict(plan, input_shape, init_model_path):
    """Initialize and return the tensor dictionary.

    Args:
        plan: The federation plan object
        input_shape: The input shape to the model
        init_model_path: Path to initial model protobuf file

    Returns:
        Tuple of (tensor_dict, task_runner, round_number)
    """
    data_loader = get_dataloader(plan, prefer_minimal=True, input_shape=input_shape)
    task_runner = plan.get_task_runner(data_loader)
    tensor_pipe = plan.get_tensor_pipe()
    round_number = 0

    if init_model_path and isfile(init_model_path):
        logger.info(f"Loading initial model from {init_model_path}")
        model_proto = utils.load_proto(init_model_path)
        init_tensor_dict, round_number = utils.deconstruct_model_proto(model_proto, tensor_pipe)
    else:
        init_tensor_dict = task_runner.get_tensor_dict(False)

    return init_tensor_dict, task_runner, round_number


# TODO: looks like Plan.method
def freeze_plan(plan_config):
    """Dump the plan to YAML file.

    Args:
        plan_config (str): Federated learning plan.
    """

    plan = Plan()
    plan.config = Plan.parse(Path(plan_config), resolve=False).config

    init_state_path = plan.config["aggregator"]["settings"]["init_state_path"]

    if not Path(init_state_path).exists():
        logger.info("Plan has not been initialized! Run 'fx plan initialize' before proceeding")
        return

    Plan.dump(Path(plan_config), plan.config, freeze=True)


@plan.command(name="freeze")
@option(
    "-p",
    "--plan_config",
    required=False,
    help="Path to an FL plan.",
    default="plan/plan.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
def freeze(plan_config):
    """Finalize the Data Science plan.

    Create a new plan file that embeds its hash in the file name
    (plan.yaml -> plan_{hash}.yaml) and changes the permissions to read only.

    Args:
        plan_config (str): Federated learning plan.
    """
    if is_directory_traversal(plan_config):
        echo("Plan config path is out of the openfl workspace scope.")
        sys.exit(1)
    freeze_plan(plan_config)


def switch_plan(name):
    """Switch the FL plan to this one.

    Args:
        name (str): Name of the Federated learning plan.
    """

    plan_file = f"plan/plans/{name}/plan.yaml"
    if isfile(plan_file):
        echo(f"Switch plan to {name}")

        # Copy the new plan.yaml file to the top directory
        copyfile(plan_file, "plan/plan.yaml")

        # Update the .workspace file to show the current workspace plan
        workspace_file = ".workspace"

        with open(workspace_file, "r", encoding="utf-8") as f:
            doc = load(f, Loader=FullLoader)

        if not doc:  # YAML is not correctly formatted
            doc = {}  # Create empty dictionary

        doc["current_plan_name"] = f"{name}"  # Switch with new plan name

        # Rewrite updated workspace file
        with open(workspace_file, "w", encoding="utf-8") as f:
            dump(doc, f)

    else:
        echo(f"Error: Plan {name} not found in plan/plans/{name}")


@plan.command(name="switch")
@option(
    "-n",
    "--name",
    required=False,
    help="Name of the FL plan to switch to.",
    default="default",
    type=str,
)
def switch_(name):
    """Switch the current plan to this plan."""
    switch_plan(name)


@plan.command(name="save")
@option(
    "-n",
    "--name",
    required=False,
    help="Name of the FL plan.",
    default="default",
    type=str,
)
def save_(name):
    """Saves the given plan and switches to it."""

    echo(f"Saving plan to {name}")
    # TODO: How do we get the prefix path? What happens if this gets executed
    #  outside of the workspace top directory?

    makedirs(f"plan/plans/{name}", exist_ok=True)
    copyfile("plan/plan.yaml", f"plan/plans/{name}/plan.yaml")

    switch_plan(name)  # Switch the context


@plan.command(name="remove")
@option(
    "-n",
    "--name",
    required=False,
    help="Name of the FL plan to remove.",
    default="default",
    type=str,
    show_default=True,
)
def remove_(name):
    """Removes given plan."""

    if name != "default":
        echo(f"Removing plan {name}")
        # TODO: How do we get the prefix path? What happens if
        #  this gets executed outside of the workspace top directory?

        rmtree(f"plan/plans/{name}")

        switch_plan("default")  # Switch the context back to the default

    else:
        echo("ERROR: Can't remove default plan")


@plan.command(name="show")
def show_():
    """Shows the active plan."""
    current_plan_name = get_workspace_parameter("current_plan_name")
    echo(f"The current plan is: {current_plan_name}")



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregator module."""

import logging
import queue
from threading import Lock
from typing import List, Optional

import openfl.callbacks as callbacks_module
from openfl.component.aggregator.straggler_handling import CutoffTimePolicy, StragglerPolicy
from openfl.databases import PersistentTensorDB, TensorDB
from openfl.pipelines import NoCompressionPipeline, TensorCodec
from openfl.protocols import base_pb2, utils
from openfl.utilities import TaskResultKey

from openfl.component import Aggregator

logger = logging.getLogger(__name__)

class AggregatorFlower(Aggregator):
    def __init__(
        self,
        aggregator_uuid,
        federation_uuid,
        authorized_cols,
        init_state_path,
        best_state_path,
        last_state_path,
        assigner,
        connector,
        use_delta_updates=True,
        straggler_handling_policy: StragglerPolicy = CutoffTimePolicy,
        rounds_to_train=256,
        single_col_cert_common_name=None,
        compression_pipeline=None,
        db_store_rounds=1,
        initial_tensor_dict=None,
        log_memory_usage=False,
        write_logs=False,
        callbacks: Optional[List] = [],
        persist_checkpoint=True,
        persistent_db_path=None,
        secure_aggregation=False,
    ):
        self.round_number = 0
        self.next_model_round_number = 0

        if single_col_cert_common_name:
            logger.warning(
                "You are running in single collaborator certificate mode. "
                "This mode is intended for development settings only and does not "
                "provide proper Public Key Infrastructure (PKI) security. "
                "Please use this mode with caution."
            )
        # FIXME: "" instead of None is for protobuf compatibility.
        self.single_col_cert_common_name = single_col_cert_common_name or ""

        self.straggler_handling_policy = straggler_handling_policy()

        self.rounds_to_train = rounds_to_train
        self.assigner = assigner
        if self.assigner.is_task_group_evaluation():
            self.rounds_to_train = 1
            logger.info(f"For evaluation tasks setting rounds_to_train = {self.rounds_to_train}")

        self._end_of_round_check_done = [False] * rounds_to_train
        self.stragglers = []

        # if the collaborator requests a delta, this value is set to true
        self.authorized_cols = authorized_cols
        self.uuid = aggregator_uuid
        self.federation_uuid = federation_uuid
        self.connector = connector

        self.quit_job_sent_to = []

        self.tensor_db = TensorDB()
        if persist_checkpoint:
            persistent_db_path = persistent_db_path or "tensor.db"
            logger.info(
                "Persistent checkpoint is enabled, setting persistent db at path %s",
                persistent_db_path,
            )
            self.persistent_db = PersistentTensorDB(persistent_db_path)
        else:
            logger.info("Persistent checkpoint is disabled")
            self.persistent_db = None
        # FIXME: I think next line generates an error on the second round
        # if it is set to 1 for the aggregator.
        self.db_store_rounds = db_store_rounds

        self.best_model_score = None
        self.metric_queue = queue.Queue()

        self.compression_pipeline = compression_pipeline or NoCompressionPipeline()
        self.tensor_codec = TensorCodec(self.compression_pipeline)

        self.init_state_path = init_state_path
        self.best_state_path = best_state_path
        self.last_state_path = last_state_path

        # TODO: Remove. Used in deprecated interactive and native APIs
        self.best_tensor_dict: dict = {}
        self.last_tensor_dict: dict = {}
        # these enable getting all tensors for a task
        self.collaborator_tasks_results = {}  # {TaskResultKey: list of TensorKeys}
        self.collaborator_task_weight = {}  # {TaskResultKey: data_size}

        # maintain a list of collaborators that have completed task and
        # reported results in a given round
        self.collaborators_done = []
        # Initialize a lock for thread safety
        self.lock = Lock()
        self.use_delta_updates = use_delta_updates

        self.model = None  # Initialize the model attribute to None

        # Callbacks
        self.callbacks = callbacks_module.CallbackList(
            callbacks,
            add_memory_profiler=log_memory_usage,
            add_metric_writer=write_logs,
            origin="aggregator",
        )

        self.collaborator_tensor_results = {}  # {TensorKey: nparray}}

        if initial_tensor_dict:
            self._load_initial_tensors_from_dict(initial_tensor_dict)
            self.model = utils.construct_model_proto(
                tensor_dict=initial_tensor_dict,
                round_number=0,
                tensor_pipe=self.compression_pipeline,
            )
        else:
            if self.connector:
                # The model definition will be handled by the respective framework
                self.model = {}
            else:
                self.model: base_pb2.ModelProto = utils.load_proto(self.init_state_path)
                self._load_initial_tensors()  # keys are TensorKeys

        self._secure_aggregation_enabled = secure_aggregation
        if self._secure_aggregation_enabled:
            from openfl.utilities.secagg.bootstrap import SecAggSetup

            self.secagg = SecAggSetup(self.uuid, self.authorized_cols, self.tensor_db)

        if self.persistent_db and self._recover():
            logger.info("Recovered state of aggregator")

        # TODO: Aggregator has no concrete notion of round_begin.
        # https://github.com/securefederatedai/openfl/pull/1195#discussion_r1879479537
        self.callbacks.on_experiment_begin()
        self.callbacks.on_round_begin(self.round_number)

    def process_task_results(
        self,
        collaborator_name,
        round_number,
        task_name,
        data_size,
        named_tensors,
    ):
        if self._time_to_quit() or collaborator_name in self.stragglers:
            logger.warning(
                f"STRAGGLER: Collaborator {collaborator_name} is reporting results "
                f"after task {task_name} has finished."
            )
            return

        if self.round_number != round_number:
            logger.warning(
                f"Collaborator {collaborator_name} is reporting results"
                f" for the wrong round: {round_number}. Ignoring..."
            )
            return

        if self.is_connector_available():
            # Skip to end of round check
            with self.lock:
                self._is_collaborator_done(collaborator_name, round_number)
                self._end_of_round_with_stragglers_check()

        task_key = TaskResultKey(task_name, collaborator_name, round_number)

        # we mustn't have results already
        if self._collaborator_task_completed(collaborator_name, task_name, round_number):
            logger.warning(
                f"Aggregator already has task results from collaborator {collaborator_name}"
                f" for task {task_key}"
            )
            return

        # By giving task_key it's own weight, we can support different
        # training/validation weights
        # As well as eventually supporting weights that change by round
        # (if more data is added)
        self.collaborator_task_weight[task_key] = data_size

        # initialize the list of tensors that go with this task
        # Setting these incrementally is leading to missing values
        task_results = []

        for named_tensor in named_tensors:
            # quite a bit happens in here, including decompression, delta
            # handling, etc...
            tensor_key, value = self._process_named_tensor(named_tensor, collaborator_name)

            if "metric" in tensor_key.tags:
                # Caution: This schema must be followed. It is also used in
                # gRPC message streams for director/envoy.
                metrics = {
                    "round": round_number,
                    "metric_origin": collaborator_name,
                    "task_name": task_name,
                    "metric_name": tensor_key.tensor_name,
                    "metric_value": float(value),
                }
                self.metric_queue.put(metrics)

            task_results.append(tensor_key)

        self.collaborator_tasks_results[task_key] = task_results

        with self.lock:
            self._is_collaborator_done(collaborator_name, round_number)

            self._end_of_round_with_stragglers_check()

    def is_connector_available(self):
        """
        Check if the OpenFL Connector is available.

        Returns:
            bool: True if connector is available, False otherwise.
        """
        return self.connector is not None

    def start_connector(self):
        """
        Start the OpenFL Connector.

        Raises:
            RuntimeError: If OpenFL Connector has not been enabled.
        """
        if not self.is_connector_available():
            raise RuntimeError("OpenFL Connector has not been enabled.")
        return self.connector.start()

    def stop_connector(self):
        """
        Stop the OpenFL Connector.

        Raises:
            RuntimeError: If OpenFL Connector has not been enabled.
        """
        if not self.is_connector_available():
            raise RuntimeError("OpenFL Connector has not been enabled.")
        return self.connector.stop()

    def get_interop_client(self):
        """
        Get the local gRPC client for the OpenFL Connector.

        Raises:
            RuntimeError: If OpenFL Connector has not been enabled.
        """
        if not self.is_connector_available():
            raise RuntimeError("OpenFL Connector has not been enabled.")
        return self.connector.get_interop_client()

    def _end_of_round_check(self):
        """Check if the round complete.

        If so, perform many end of round operations,
        such as model aggregation, metric reporting, delta generation (+
        associated tensorkey labeling), and save the model.

        Args:
            None

        Returns:
            None
        """
        if self._end_of_round_check_done[self.round_number]:
            return

        if not self.is_connector_available():
        # Compute all validation related metrics
            logs = {}
            for task_name in self.assigner.get_all_tasks_for_round(self.round_number):
                logs.update(self._compute_validation_related_task_metrics(task_name))

            # End of round callbacks.
            self.callbacks.on_round_end(self.round_number, logs)

        # Once all of the task results have been processed
        self._end_of_round_check_done[self.round_number] = True

        # Save the latest model
        if not self.is_connector_available():
            logger.info("Saving round %s model...", self.round_number)
            self._save_model(self.round_number, self.last_state_path)

        self.round_number += 1
        # resetting stragglers for task for a new round
        self.stragglers = []
        # resetting collaborators_done for next round
        self.collaborators_done = []

        # TODO This needs to be fixed!
        if self._time_to_quit():
            logger.info("Experiment Completed. Cleaning up...")
        else:
            logger.info("Starting round %s...", self.round_number)
            # https://github.com/securefederatedai/openfl/pull/1195#discussion_r1879479537
            self.callbacks.on_round_begin(self.round_number)

        # Cleaning tensor db
        self.tensor_db.clean_up(self.db_store_rounds)
        # Reset straggler handling policy for the next round.
        self.straggler_handling_policy.reset_policy_for_round()



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Collaborator module."""

import os
import sys
from glob import glob
from logging import getLogger
from os import remove
from os.path import basename, isfile, join, splitext
from pathlib import Path
from shutil import copy, copytree, ignore_patterns, make_archive, unpack_archive
from tempfile import mkdtemp

import yaml
from click import Path as ClickPath
from click import confirm, echo, group, option, pass_context, style
from yaml import FullLoader, dump, load

from openfl.cryptography.ca import sign_certificate
from openfl.cryptography.io import get_csr_hash, read_crt, read_csr, read_key, write_crt, write_key
from openfl.cryptography.participant import generate_csr
from openfl.experimental.workflow.federated import Plan
from openfl.experimental.workflow.interface.cli.cli_helper import CERT_DIR
from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.utils import rmtree

logger = getLogger(__name__)


@group()
@pass_context
def collaborator(context):
    """Manage Federated Learning Collaborators."""
    context.obj["group"] = "service"


@collaborator.command(name="start")
@option(
    "-p",
    "--plan",
    required=False,
    help="Federated learning plan [plan/plan.yaml]",
    default="plan/plan.yaml",
    type=ClickPath(exists=True),
)
@option(
    "-n",
    "--collaborator_name",
    required=True,
    help="The certified common name of the collaborator",
)
@option(
    "-s",
    "--secure",
    required=False,
    help="Enable Intel SGX Enclave",
    is_flag=True,
    default=False,
)
def start_(plan, collaborator_name, secure, data_config="plan/data.yaml"):
    """Start a collaborator service."""

    if plan and is_directory_traversal(plan):
        echo("Federated learning plan path is out of the openfl workspace scope.")
        sys.exit(1)
    if data_config and is_directory_traversal(data_config):
        echo("The data set/shard configuration file path is out of the openfl workspace scope.")
        sys.exit(1)

    plan = Plan.parse(
        plan_config_path=Path(plan).absolute(),
        data_config_path=Path(data_config).absolute(),
    )

    if not os.path.exists(data_config):
        logger.warning(
            "Collaborator private attributes are set to None as"
            f" {data_config} not found in workspace."
        )
    else:
        with open(data_config, "r") as f:
            data = yaml.safe_load(f)
            if data.get(collaborator_name, None) is None:
                logger.warning(
                    f"Collaborator private attributes are set to None as no attributes"
                    f" for {collaborator_name} found in {data_config}."
                )

    logger.info("🧿 Starting the Collaborator Service.")

    plan.get_collaborator(collaborator_name).run()


@collaborator.command(name="generate-cert-request")
@option(
    "-n",
    "--collaborator_name",
    required=True,
    help="The certified common name of the collaborator",
)
@option("-s", "--silent", help="Do not prompt", is_flag=True)
@option(
    "-x",
    "--skip-package",
    help="Do not package the certificate signing request for export",
    is_flag=True,
)
def generate_cert_request_(collaborator_name, silent, skip_package):
    """Generate certificate request for the collaborator."""
    generate_cert_request(collaborator_name, silent, skip_package)


def generate_cert_request(collaborator_name, silent, skip_package):
    """Create collaborator certificate key pair.

    Then create a package with the CSR to send for signing.
    """

    common_name = f"{collaborator_name}"
    subject_alternative_name = f"DNS:{common_name}"
    file_name = f"col_{common_name}"

    echo(
        f"Creating COLLABORATOR certificate key pair with following settings: "
        f"CN={style(common_name, fg='red')},"
        f" SAN={style(subject_alternative_name, fg='red')}"
    )

    client_private_key, client_csr = generate_csr(common_name, server=False)

    (CERT_DIR / "client").mkdir(parents=True, exist_ok=True)

    echo("  Moving COLLABORATOR certificate to: " + style(f"{CERT_DIR}/{file_name}", fg="green"))

    # Print csr hash before writing csr to disk
    csr_hash = get_csr_hash(client_csr)
    echo("The CSR Hash " + style(f"{csr_hash}", fg="red"))

    # Write collaborator csr and key to disk
    write_crt(client_csr, CERT_DIR / "client" / f"{file_name}.csr")
    write_key(client_private_key, CERT_DIR / "client" / f"{file_name}.key")

    if not skip_package:
        archive_type = "zip"
        archive_name = f"col_{common_name}_to_agg_cert_request"
        archive_file_name = archive_name + "." + archive_type

        # Collaborator certificate signing request
        tmp_dir = join(mkdtemp(), "openfl", archive_name)

        ignore = ignore_patterns("__pycache__", "*.key", "*.srl", "*.pem")
        # Copy the current directory into the temporary directory
        copytree(f"{CERT_DIR}/client", tmp_dir, ignore=ignore)

        for f in glob(f"{tmp_dir}/*"):
            if common_name not in basename(f):
                remove(f)

        # Create Zip archive of directory
        make_archive(archive_name, archive_type, tmp_dir)
        rmtree(tmp_dir)

        echo(f"Archive {archive_file_name} with certificate signing request created")
        echo(
            "This file should be sent to the certificate authority"
            " (typically hosted by the aggregator) for signing"
        )


def find_certificate_name(file_name):
    """Parse the collaborator name."""
    col_name = str(file_name).split(os.sep)[-1].split(".")[0][4:]
    return col_name


def register_collaborator(file_name):
    """Register the collaborator name in the cols.yaml list.

    Args:
        file_name (str): The name of the collaborator in this federation
    """

    col_name = find_certificate_name(file_name)

    cols_file = Path("plan/cols.yaml").absolute()

    if not isfile(cols_file):
        cols_file.touch()
    with open(cols_file, "r", encoding="utf-8") as f:
        doc = load(f, Loader=FullLoader)

    if not doc:  # YAML is not correctly formatted
        doc = {}  # Create empty dictionary

    # List doesn't exist
    if "collaborators" not in doc.keys() or not doc["collaborators"]:
        doc["collaborators"] = []  # Create empty list

    if col_name in doc["collaborators"]:
        echo(
            "\nCollaborator "
            + style(f"{col_name}", fg="green")
            + " is already in the "
            + style(f"{cols_file}", fg="green")
        )

    else:
        doc["collaborators"].append(col_name)
        with open(cols_file, "w", encoding="utf-8") as f:
            dump(doc, f)

        echo(
            "\nRegistering "
            + style(f"{col_name}", fg="green")
            + " in "
            + style(f"{cols_file}", fg="green")
        )


@collaborator.command(name="certify")
@option(
    "-n",
    "--collaborator_name",
    help="The certified common name of the collaborator. This is only"
    " needed for single node experiments",
)
@option("-s", "--silent", help="Do not prompt", is_flag=True)
@option(
    "-r",
    "--request-pkg",
    type=ClickPath(exists=True),
    help="The archive containing the certificate signing request (*.zip) for a collaborator",
)
@option(
    "-i",
    "--import",
    "import_",
    type=ClickPath(exists=True),
    help="Import the archive containing the collaborator's certificate (signed by the CA)",
)
def certify_(collaborator_name, silent, request_pkg, import_):
    """Certify the collaborator."""
    certify(collaborator_name, silent, request_pkg, import_)


def _handle_request_package(request_pkg):
    Path(f"{CERT_DIR}/client").mkdir(parents=True, exist_ok=True)
    unpack_archive(request_pkg, extract_dir=f"{CERT_DIR}/client")
    return glob(f"{CERT_DIR}/client/*.csr")[0]


def _load_certificate_files(common_name, request_pkg):
    if request_pkg:
        return _handle_request_package(request_pkg)

    csr_path = glob(f"{CERT_DIR}/client/col_{common_name}.csr")[0]
    copy(csr_path, CERT_DIR)
    return csr_path


def _validate_paths(cert_name):
    signing_key_path = CERT_DIR / "ca/signing-ca/private/signing-ca.key"
    signing_crt_path = CERT_DIR / "ca/signing-ca.crt"

    for path, error_message in [
        (cert_name + ".csr", "Collaborator certificate signing request not found."),
        (signing_key_path, "Signing key not found."),
        (signing_crt_path, "Signing certificate not found."),
    ]:
        if not Path(path).exists():
            echo(style(error_message, fg="red"))
            return None, None

    return read_key(signing_key_path), read_crt(signing_crt_path)


def _sign_and_register_certificate(file_name, csr, csr_hash, signing_key, signing_crt, silent):
    echo(
        f"The CSR Hash for file {style(f'{file_name}.csr', fg='green')} = "
        f"{style(f'{csr_hash}', fg='red')}"
    )

    if silent:
        echo(
            "Signing COLLABORATOR certificate\n"
            "Warning: manual check of certificate hashes is bypassed in silent mode."
        )
    else:
        echo("Make sure the two hashes above are the same.")
        if not confirm("Do you want to sign this certificate?"):
            echo(
                style("Not signing certificate.", fg="red")
                + " Please check with this collaborator to get the"
                " correct certificate for this federation."
            )
            return

    signed_cert = sign_certificate(csr, signing_key, signing_crt.subject)
    write_crt(signed_cert, f"{CERT_DIR}/client/{file_name}.crt")
    register_collaborator(CERT_DIR / "client" / f"{file_name}.crt")


def _package_signed_certificate(file_name):
    archive_type = "zip"
    archive_name = f"agg_to_{file_name}_signed_cert"
    tmp_dir = join(mkdtemp(), "openfl", archive_name)

    Path(f"{tmp_dir}/client").mkdir(parents=True, exist_ok=True)
    # Copy the signed cert to the temporary directory
    copy(f"{CERT_DIR}/client/{file_name}.crt", f"{tmp_dir}/client/")
    # Copy the CA certificate chain to the temporary directory
    copy(f"{CERT_DIR}/cert_chain.crt", tmp_dir)

    # Create Zip archive of directory
    make_archive(archive_name, archive_type, tmp_dir)
    rmtree(tmp_dir)


def _import_signed_certificate(import_):
    # Copy the signed certificate and cert chain into PKI_DIR
    previous_crts = glob(f"{CERT_DIR}/client/*.crt")
    unpack_archive(import_, extract_dir=CERT_DIR)
    updated_crts = glob(f"{CERT_DIR}/client/*.crt")
    cert_difference = list(set(updated_crts) - set(previous_crts))

    if cert_difference:
        echo(f"Certificate {basename(cert_difference[0])} installed to PKI directory")
    else:
        echo("Certificate updated in the PKI directory")


def certify(collaborator_name, silent, request_pkg=None, import_=False):
    if import_:
        return _import_signed_certificate(import_)

    if request_pkg is None and collaborator_name is None:
        echo(
            "collaborator_name can only be omitted if signing\n"
            "a zipped request package.\n"
            "\n"
            "Example: fx collaborator certify --request-pkg "
            "col_one_to_agg_cert_request.zip"
        )
        return

    common_name = f"{collaborator_name}"
    if len(common_name) == 0:
        # If the collaborator name is provided, the collaborator and
        # certificate does not need to be exported
        return

    csr_path = _load_certificate_files(common_name, request_pkg)
    cert_name = splitext(csr_path)[0]
    file_name = basename(cert_name)

    signing_key, signing_crt = _validate_paths(cert_name)
    if not signing_key or not signing_crt:
        return

    csr, csr_hash = read_csr(csr_path)
    _sign_and_register_certificate(file_name, csr, csr_hash, signing_key, signing_crt, silent)

    remove(csr_path)
    _package_signed_certificate(file_name)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Experimental Envoy module."""

import logging
import sys
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Optional, Union

from openfl.experimental.workflow.federated import Plan
from openfl.experimental.workflow.transport.grpc.director_client import EnvoyDirectorClient
from openfl.experimental.workflow.transport.grpc.exceptions import EnvoyNotFoundError
from openfl.utilities.workspace import ExperimentWorkspace

logger = logging.getLogger(__name__)


class Envoy:
    """Envoy class. The Envoy is a long-lived entity that runs on collaborator
    nodes connected to the Director.

    Attributes:
        name (str): The name of the envoy.
        envoy_config (Optional[Path]): Path to envoy_config.yaml
        tls (bool, optional): A flag indicating if TLS should be used for
                connections. Defaults to True.
        root_certificate (Optional[Union[Path, str]]): The path to the root certificate
            for TLS.
        private_key (Optional[Union[Path, str]]): The path to the private key for TLS.
        certificate (Optional[Union[Path, str]]): The path to the certificate for TLS.
        _envoy_dir_client (EnvoyDirectorClient): The envoy director client.
        install_requirements (bool): A flag indicating if the requirements
            should be installed.
        is_experiment_running (bool): A flag indicating if an experiment is
            running.
        executor (ThreadPoolExecutor): The executor for running tasks.
        plan(str): Path to plan.yaml
        _health_check_future (object): The future object for the health check.
    """

    DEFAULT_RETRY_TIMEOUT_IN_SECONDS = 5

    def __init__(
        self,
        *,
        envoy_name: str,
        director_host: str,
        director_port: int,
        envoy_config: Optional[Path] = None,
        root_certificate: Optional[Union[Path, str]] = None,
        private_key: Optional[Union[Path, str]] = None,
        certificate: Optional[Union[Path, str]] = None,
        tls: bool = True,
        install_requirements: bool = True,
    ) -> None:
        """Initialize a envoy object.

        Args:
            envoy_name (str): The name of the envoy.
            director_host (str): The host of the director.
            director_port (int): The port of the director.
            envoy_config (Optional[Path]): Path to envoy_config.yaml
            root_certificate (Optional[Union[Path, str]]): The path
                to the root certificate for TLS. Defaults to None.
            private_key (Optional[Union[Path, str]]): The path to
                the private key for TLS. Defaults to None.
            certificate (Optional[Union[Path, str]]): The path to
                the certificate for TLS. Defaults to None.
            tls (bool, optional): A flag indicating if TLS should be used for
                connections. Defaults to True.
            install_requirements (bool, optional): A flag indicating if the
                requirements should be installed. Defaults to True.
        """
        self.name = envoy_name
        self.envoy_config = envoy_config
        self.tls = tls
        self._fill_certs(root_certificate, private_key, certificate)
        self.install_requirements = install_requirements
        self._envoy_dir_client = self._create_envoy_dir_client(director_host, director_port)
        self.is_experiment_running = False
        self.executor = ThreadPoolExecutor()
        # This plan path ("plan/plan.yaml") originates from the
        # experiment workspace provided by the director
        self.plan = "plan/plan.yaml"
        self._health_check_future = None

    def _create_envoy_dir_client(
        self, director_host: str, director_port: int
    ) -> EnvoyDirectorClient:
        """Create a EnvoyDirectorClient instance.

        Args:
            director_host (str): The host of the director.
            director_port (int): The port of the director.

        Returns:
            EnvoyDirectorClient: Instance of the client
        """
        return EnvoyDirectorClient(
            director_host=director_host,
            director_port=director_port,
            envoy_name=self.name,
            tls=self.tls,
            root_certificate=self.root_certificate,
            private_key=self.private_key,
            certificate=self.certificate,
        )

    def _fill_certs(self, root_certificate, private_key, certificate) -> None:
        """Fill certificates.

        Args:
            root_certificate (Union[Path, str]): The path to the root
                certificate for the TLS connection.
            private_key (Union[Path, str]): The path to the server's private
                key for the TLS connection.
            certificate (Union[Path, str]): The path to the server's
                certificate for the TLS connection.
        """
        if self.tls:
            if not all([root_certificate, private_key, certificate]):
                raise ValueError("Incomplete certificates provided")

            self.root_certificate = Path(root_certificate).absolute()
            self.private_key = Path(private_key).absolute()
            self.certificate = Path(certificate).absolute()
        else:
            self.root_certificate = self.private_key = self.certificate = None

    def _run(self) -> None:
        """Run of the envoy working cycle."""
        while True:
            try:
                # Wait for experiment from Director server
                experiment_name = self._envoy_dir_client.wait_experiment()
                data_stream = self._envoy_dir_client.get_experiment_data(experiment_name)
            except Exception as exc:
                logger.exception("Failed to get experiment: %s", exc)
                time.sleep(self.DEFAULT_RETRY_TIMEOUT_IN_SECONDS)
                continue
            data_file_path = self._save_data_stream_to_file(data_stream)

            try:
                with ExperimentWorkspace(
                    experiment_name=f"{self.name}_{experiment_name}",
                    data_file_path=data_file_path,
                    install_requirements=self.install_requirements,
                ):
                    self.is_experiment_running = True
                    self._run_collaborator()
            except Exception as exc:
                logger.exception("Collaborator failed with error: %s:", exc)
            finally:
                self.is_experiment_running = False

    @staticmethod
    def _save_data_stream_to_file(data_stream) -> Path:
        """Save data stream to file.

        Args:
            data_stream: The data stream to save.

        Returns:
            Path: The path to the saved data file.
        """
        data_file_path = Path(str(uuid.uuid4())).absolute()
        with open(data_file_path, "wb") as data_file:
            for response in data_stream:
                if response.size == len(response.exp_data):
                    data_file.write(response.exp_data)
                else:
                    raise Exception("Broken archive")
        return data_file_path

    def _send_health_check(self) -> None:
        """Send health check to the director."""
        logger.debug("Sending envoy node status to director.")
        timeout = self.DEFAULT_RETRY_TIMEOUT_IN_SECONDS
        while True:
            try:
                timeout = self._envoy_dir_client.send_health_check(
                    envoy_name=self.name,
                    is_experiment_running=self.is_experiment_running,
                )
            except EnvoyNotFoundError:
                logger.info(
                    "The director has lost information about current envoy. Reconnecting..."
                )
                self._envoy_dir_client.connect_envoy(envoy_name=self.name)
            time.sleep(timeout)

    def _run_collaborator(self) -> None:
        """Run the collaborator for the experiment running."""
        plan = Plan.parse(plan_config_path=Path(self.plan))
        logger.info("🧿 Starting the Collaborator Service.")

        col = plan.get_collaborator(
            self.name,
            self.root_certificate,
            self.private_key,
            self.certificate,
            envoy_config=self.envoy_config,
            tls=self.tls,
        )
        col.run()

    def start(self) -> None:
        """Start the envoy"""
        try:
            is_accepted = self._envoy_dir_client.connect_envoy(envoy_name=self.name)
        except Exception as exc:
            logger.exception("Failed to connect envoy: %s", exc)
            sys.exit(1)
        else:
            if is_accepted:
                logger.info(f"{self.name} is connected to the director")
                self._health_check_future = self.executor.submit(self._send_health_check)
                self._run()
            else:
                # Connection failed
                logger.error(f"{self.name} failed to connect to the director")
                sys.exit(1)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Workspace module."""

import os
import sys
from hashlib import sha256
from logging import getLogger
from os import chdir, getcwd, makedirs
from os.path import basename, isfile, join
from pathlib import Path
from shutil import copy2, copyfile, copytree, ignore_patterns, make_archive, unpack_archive
from subprocess import check_call
from sys import executable
from tempfile import mkdtemp
from typing import Tuple

from click import Choice, confirm, echo, group, option, pass_context, style
from click import Path as ClickPath
from cryptography.hazmat.primitives import serialization

from openfl.cryptography.ca import generate_root_cert, generate_signing_csr, sign_certificate
from openfl.experimental.workflow.federated.plan import Plan
from openfl.experimental.workflow.interface.cli.cli_helper import (
    CERT_DIR,
    OPENFL_USERDIR,
    WORKSPACE,
    print_tree,
)
from openfl.experimental.workflow.interface.cli.plan import freeze_plan
from openfl.experimental.workflow.notebooktools import NotebookTools
from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.utils import rmtree
from openfl.utilities.workspace import dump_requirements_file

logger = getLogger(__name__)


@group()
@pass_context
def workspace(context):
    """Manage Experimental Federated Learning Workspaces."""
    context.obj["group"] = "workspace"


def create_dirs(prefix):
    """Create workspace directories."""

    echo("Creating Workspace Directories")

    (prefix / "cert").mkdir(parents=True, exist_ok=True)  # certifications
    (prefix / "data").mkdir(parents=True, exist_ok=True)  # training data
    (prefix / "logs").mkdir(parents=True, exist_ok=True)  # training logs
    (prefix / "save").mkdir(parents=True, exist_ok=True)  # model weight saves / initialization
    (prefix / "src").mkdir(parents=True, exist_ok=True)  # model code

    copyfile(WORKSPACE / "workspace" / ".workspace", prefix / ".workspace")


def create_temp(prefix, template):
    """Create workspace templates."""

    echo("Creating Workspace Templates")
    # Use the specified template if it's a Path, otherwise use WORKSPACE/template
    source = template if isinstance(template, Path) else WORKSPACE / template

    copytree(
        src=source,
        dst=prefix,
        dirs_exist_ok=True,
        ignore=ignore_patterns("__pycache__"),
    )  # from template workspace
    apply_template_plan(prefix, template)


def get_templates():
    """Grab the default templates from the distribution."""

    return [
        d.name
        for d in WORKSPACE.glob("*")
        if d.is_dir() and d.name not in ["__pycache__", "workspace"]
    ]


@workspace.command(name="create")
@option("--prefix", required=True, help="Workspace name or path", type=ClickPath())
@option(
    "--custom_template",
    required=False,
    help="Path to custom template",
    type=ClickPath(exists=True),
)
@option(
    "--notebook",
    required=False,
    help="Path to jupyter notebook",
    type=ClickPath(exists=True),
)
@option(
    "--template_output_dir",
    required=False,
    help="Destination directory to save your Jupyter Notebook workspace.",
    type=ClickPath(exists=False, file_okay=False, dir_okay=True),
)
@option("--template", required=False, type=Choice(get_templates()))
def create_(prefix, custom_template, template, notebook, template_output_dir):
    """Create the experimental workspace."""
    if is_directory_traversal(prefix):
        echo("Workspace name or path is out of the openfl workspace scope.")
        sys.exit(1)

    if custom_template and template and notebook:
        raise ValueError(
            "Please provide either `template`, `custom_template` or "
            + "`notebook`. Not all are necessary"
        )
    elif (
        (custom_template and template) or (template and notebook) or (custom_template and notebook)
    ):
        raise ValueError(
            "Please provide only one of the following options: "
            + "`template`, `custom_template`, or `notebook`."
        )

    if not (custom_template or template or notebook):
        raise ValueError(
            "Please provide one of the following options: "
            + "`template`, `custom_template`, or `notebook`."
        )
    # FIXME: Remove this CLI Aggregator-Based functionality in the subsequent PR
    if notebook:
        if not template_output_dir:
            raise ValueError(
                "Please provide output_workspace which is Destination directory to "
                + "save your Jupyter Notebook workspace."
            )

        NotebookTools.export(
            notebook_path=notebook,
            output_workspace=template_output_dir,
        )

        create(prefix, Path(template_output_dir).resolve())

        logger.warning(
            "The user should review the generated workspace for completeness " + "before proceeding"
        )
    else:
        template = Path(custom_template).resolve() if custom_template else template
        create(prefix, template)


def create(prefix, template):
    """Create federated learning workspace."""

    if not OPENFL_USERDIR.exists():
        OPENFL_USERDIR.mkdir()

    prefix = Path(prefix).absolute()

    create_dirs(prefix)
    create_temp(prefix, template)

    requirements_filename = "requirements.txt"

    if not os.path.exists(f"{str(prefix)}/plan/data.yaml"):
        echo(
            style(
                "Participant private attributes shall be set to None as plan/data.yaml"
                + " was not found in the workspace.",
                fg="yellow",
            )
        )

    if isfile(f"{str(prefix)}/{requirements_filename}"):
        check_call(
            [
                executable,
                "-m",
                "pip",
                "install",
                "-r",
                f"{prefix}/requirements.txt",
            ],
            shell=False,
        )
        echo(f"Successfully installed packages from {prefix}/requirements.txt.")
    else:
        echo("No additional requirements for workspace defined. Skipping...")
    prefix_hash = _get_dir_hash(str(prefix.absolute()))
    with open(
        OPENFL_USERDIR / f"requirements.{prefix_hash}.txt",
        "w",
        encoding="utf-8",
    ) as f:
        check_call([executable, "-m", "pip", "freeze"], shell=False, stdout=f)

    print_tree(prefix, level=3)


@workspace.command(name="export")
@option(
    "-o",
    "--pip-install-options",
    required=False,
    type=str,
    multiple=True,
    default=tuple,
    help="Options for remote pip install. "
    "You may pass several options in quotation marks alongside with arguments, "
    'e.g. -o "--find-links source.site"',
)
def export_(pip_install_options: Tuple[str]):
    """Export federated learning workspace."""

    echo(
        style(
            "This command will archive the contents of 'plan' and 'src' directory, user"
            + " should review that these does not contain any information which is private and"
            + " not to be shared.",
            fg="yellow",
        )
    )

    plan_file = Path("plan/plan.yaml").absolute()
    try:
        freeze_plan(plan_file)
    except FileNotFoundError:
        echo(f'Plan file "{plan_file}" not found. No freeze performed.')

    # Dump requirements.txt
    dump_requirements_file(prefixes=pip_install_options, keep_original_prefixes=True)

    archive_type = "zip"
    archive_name = basename(getcwd())
    archive_file_name = archive_name + "." + archive_type

    # Aggregator workspace
    tmp_dir = join(mkdtemp(), "openfl", archive_name)

    ignore = ignore_patterns("__pycache__", "*.crt", "*.key", "*.csr", "*.srl", "*.pem", "*.pbuf")

    # We only export the minimum required files to set up a collaborator
    makedirs(f"{tmp_dir}/save", exist_ok=True)
    makedirs(f"{tmp_dir}/logs", exist_ok=True)
    makedirs(f"{tmp_dir}/data", exist_ok=True)
    copytree("./src", f"{tmp_dir}/src", ignore=ignore)  # code
    copytree("./plan", f"{tmp_dir}/plan", ignore=ignore)  # plan
    copy2("./requirements.txt", f"{tmp_dir}/requirements.txt")  # requirements

    try:
        copy2(".workspace", tmp_dir)  # .workspace
    except FileNotFoundError:
        echo("'.workspace' file not found.")
        if confirm("Create a default '.workspace' file?"):
            copy2(WORKSPACE / "workspace" / ".workspace", tmp_dir)
        else:
            echo("To proceed, you must have a '.workspace' file in the current directory.")
            raise

    # Create Zip archive of directory
    echo("\n 🗜️ Preparing workspace distribution zip file")
    make_archive(archive_name, archive_type, tmp_dir)
    rmtree(tmp_dir)
    echo(f"\n ✔️ Workspace exported to archive: {archive_file_name}")


@workspace.command(name="import")
@option(
    "--archive",
    required=True,
    help="Zip file containing workspace to import",
    type=ClickPath(exists=True),
)
def import_(archive):
    """Import federated learning workspace."""

    archive = Path(archive).absolute()

    dir_path = basename(archive).split(".")[0]
    unpack_archive(archive, extract_dir=dir_path)
    chdir(dir_path)

    requirements_filename = "requirements.txt"

    if isfile(requirements_filename):
        check_call(
            [executable, "-m", "pip", "install", "--upgrade", "pip"],
            shell=False,
        )
        check_call(
            [executable, "-m", "pip", "install", "-r", requirements_filename],
            shell=False,
        )
    else:
        echo("No " + requirements_filename + " file found.")

    echo(f"Workspace {archive} has been imported.")
    echo("You may need to copy your PKI certificates to join the federation.")


@workspace.command(name="certify")
def certify_():
    """Create certificate authority for federation."""
    certify()


def certify():
    """Create certificate authority for federation."""

    echo("Setting Up Certificate Authority...\n")

    echo("1.  Create Root CA")
    echo("1.1 Create Directories")

    (CERT_DIR / "ca/root-ca/private").mkdir(parents=True, exist_ok=True, mode=0o700)
    (CERT_DIR / "ca/root-ca/db").mkdir(parents=True, exist_ok=True)

    echo("1.2 Create Database")

    with open(CERT_DIR / "ca/root-ca/db/root-ca.db", "w", encoding="utf-8") as f:
        pass  # write empty file
    with open(CERT_DIR / "ca/root-ca/db/root-ca.db.attr", "w", encoding="utf-8") as f:
        pass  # write empty file

    with open(CERT_DIR / "ca/root-ca/db/root-ca.crt.srl", "w", encoding="utf-8") as f:
        f.write("01")  # write file with '01'
    with open(CERT_DIR / "ca/root-ca/db/root-ca.crl.srl", "w", encoding="utf-8") as f:
        f.write("01")  # write file with '01'

    echo("1.3 Create CA Request and Certificate")

    root_crt_path = "ca/root-ca.crt"
    root_key_path = "ca/root-ca/private/root-ca.key"

    root_private_key, root_cert = generate_root_cert()

    # Write root CA certificate to disk
    with open(CERT_DIR / root_crt_path, "wb") as f:
        f.write(
            root_cert.public_bytes(
                encoding=serialization.Encoding.PEM,
            )
        )

    with open(CERT_DIR / root_key_path, "wb") as f:
        f.write(
            root_private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    echo("2.  Create Signing Certificate")
    echo("2.1 Create Directories")

    (CERT_DIR / "ca/signing-ca/private").mkdir(parents=True, exist_ok=True, mode=0o700)
    (CERT_DIR / "ca/signing-ca/db").mkdir(parents=True, exist_ok=True)

    echo("2.2 Create Database")

    with open(CERT_DIR / "ca/signing-ca/db/signing-ca.db", "w", encoding="utf-8") as f:
        pass  # write empty file
    with open(CERT_DIR / "ca/signing-ca/db/signing-ca.db.attr", "w", encoding="utf-8") as f:
        pass  # write empty file

    with open(CERT_DIR / "ca/signing-ca/db/signing-ca.crt.srl", "w", encoding="utf-8") as f:
        f.write("01")  # write file with '01'
    with open(CERT_DIR / "ca/signing-ca/db/signing-ca.crl.srl", "w", encoding="utf-8") as f:
        f.write("01")  # write file with '01'

    echo("2.3 Create Signing Certificate CSR")

    signing_csr_path = "ca/signing-ca.csr"
    signing_crt_path = "ca/signing-ca.crt"
    signing_key_path = "ca/signing-ca/private/signing-ca.key"

    signing_private_key, signing_csr = generate_signing_csr()

    # Write Signing CA CSR to disk
    with open(CERT_DIR / signing_csr_path, "wb") as f:
        f.write(
            signing_csr.public_bytes(
                encoding=serialization.Encoding.PEM,
            )
        )

    with open(CERT_DIR / signing_key_path, "wb") as f:
        f.write(
            signing_private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    echo("2.4 Sign Signing Certificate CSR")

    signing_cert = sign_certificate(signing_csr, root_private_key, root_cert.subject, ca=True)

    with open(CERT_DIR / signing_crt_path, "wb") as f:
        f.write(
            signing_cert.public_bytes(
                encoding=serialization.Encoding.PEM,
            )
        )

    echo("3   Create Certificate Chain")

    # create certificate chain file by combining root-ca and signing-ca
    with open(CERT_DIR / "cert_chain.crt", "w", encoding="utf-8") as d:
        with open(CERT_DIR / "ca/root-ca.crt", encoding="utf-8") as s:
            d.write(s.read())
        with open(CERT_DIR / "ca/signing-ca.crt") as s:
            d.write(s.read())

    echo("\nDone.")


# FIXME: Function is not in use


def _get_requirements_dict(txtfile):
    with open(txtfile, "r", encoding="utf-8") as snapshot:
        snapshot_dict = {}
        for line in snapshot:
            try:
                # 'pip freeze' generates requirements with exact versions
                k, v = line.split("==")
                snapshot_dict[k] = v
            except ValueError:
                snapshot_dict[line] = None
        return snapshot_dict


def _get_dir_hash(path):
    hash_ = sha256()
    hash_.update(path.encode("utf-8"))
    hash_ = hash_.hexdigest()
    return hash_


def apply_template_plan(prefix, template):
    """Copy plan file from template folder.

    This function unfolds default values from template plan configuration and
    writes the configuration to the current workspace.
    """

    # Use the specified template if it's a Path, otherwise use
    # WORKSPACE/template
    source = template if isinstance(template, Path) else WORKSPACE / template

    template_plan = Plan.parse(source / "plan" / "plan.yaml")

    Plan.dump(prefix / "plan" / "plan.yaml", template_plan.config)



from openfl.federated.task.runner import TaskRunner
import subprocess
from logging import getLogger
import time
import os
import numpy as np
from pathlib import Path
import sys
import socket
from src.util import is_safe_path

flwr_home = os.path.join(os.getcwd(), "save/.flwr")
if not is_safe_path(flwr_home):
    raise ValueError("Invalid path for FLWR_HOME")

os.environ["FLWR_HOME"] = flwr_home
os.makedirs(os.environ["FLWR_HOME"], exist_ok=True)

class FlowerTaskRunner(TaskRunner):
    """
    FlowerTaskRunner is a task runner that executes the Flower SuperNode
    to initialize and manage experiments from the client side.

    This class is responsible for starting a local gRPC server and a Flower SuperNode
    in a subprocess. It provides options for both manual and automatic shutdown based on
    subprocess activity.
    """
    def __init__(self, **kwargs):
        """
        Initialize the FlowerTaskRunner.

        Args:
            **kwargs: Additional parameters to pass to the functions.
        """
        super().__init__(**kwargs)

        self.patch = kwargs.get('patch')
        if self.data_loader is None:
            flwr_app_name = kwargs.get('flwr_app_name')
            if self.patch:
                install_flower_FAB(flwr_app_name)
            return

        self.model = None
        self.logger = getLogger(__name__)

        self.data_path = self.data_loader.get_node_configs()

        self.client_port = kwargs.get('client_port')
        if self.client_port is None:
            self.client_port = get_dynamic_port()

        self.shutdown_requested = False  # Flag to signal shutdown

    def start_client_adapter(self, local_grpc_server, **kwargs):
        """
        Start the local gRPC server and the Flower SuperNode.

        Args:
            local_grpc_server: The local gRPC server instance.
            **kwargs: Additional parameters, including 'local_server_port'.
        """
        local_server_port = kwargs.get('local_server_port')

        def message_callback():
            self.shutdown_requested = True

        # Set the callback for ending the experiment
        local_grpc_server.set_end_experiment_callback(message_callback)
        local_grpc_server.start_server(local_server_port)

        local_server_port = local_grpc_server.get_port()

        command = [
            "flower-supernode",
            "--insecure",
            "--grpc-adapter",
            "--superlink", f"127.0.0.1:{local_server_port}",
            "--clientappio-api-address", f"127.0.0.1:{self.client_port}",
            "--node-config", f"data-path='{self.data_path}'"
        ]

        supernode_process = subprocess.Popen(command, shell=False)
        local_grpc_server.handle_signals(supernode_process)

        self.logger.info("Press CTRL+C to stop the server and SuperNode process.")

        try:
            while not local_grpc_server.termination_event.is_set():
                if self.shutdown_requested:
                    local_grpc_server.terminate_supernode_process(supernode_process)
                    local_grpc_server.stop_server()
                time.sleep(0.1)
        except KeyboardInterrupt:
            local_grpc_server.terminate_supernode_process(supernode_process)
            local_grpc_server.stop_server()

    def set_tensor_dict(self, tensor_dict, with_opt_vars=False):
        """
        Set the tensor dictionary for the task runner.

        This method is framework agnostic and does not attempt to load the weights into the model
        or save out the native format. Instead, it directly loads and saves the dictionary.

        Args:
            tensor_dict (dict): The tensor dictionary.
            with_opt_vars (bool): This argument is inherited from the parent class
                but is not used in the FlowerTaskRunner.
        """
        self.tensor_dict = tensor_dict

    def save_native(self, filepath, **kwargs):
        """
        Save model weights to a .npz file specified by the filepath.

        The model weights are stored as a dictionary of np.ndarray.

        Args:
            filepath (str): Path to the .npz file to be created by np.savez().
            **kwargs: Additional parameters (currently not used).

        Returns:
            None

        Raises:
            AssertionError: If the file extension is not '.npz'.
        """
        # Ensure the file extension is .npz
        if isinstance(filepath, Path):
            filepath = str(filepath)

        assert filepath.endswith('.npz'), "Currently, only '.npz' file type is supported."

        # Save the tensor dictionary to a .npz file
        np.savez(filepath, **self.tensor_dict)

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """Initialize tensor keys for functions. Currently not implemented."""
        pass


def install_flower_FAB(flwr_app_name):
    """
    Build and install the patch for the Flower application.

    Args:
        flwr_app_name (str): The name of the Flower application to patch.
    """
    flwr_dir = os.environ["FLWR_HOME"]
    os.environ["TMPDIR"] = flwr_dir

    # Run the build command
    subprocess.check_call([
        sys.executable,
        "src/patch/flwr_run_patch.py",
        "build",
        "--app",
        f"./src/{flwr_app_name}"
    ])

    # List .fab files after running the build command
    fab_files = list(Path(flwr_dir).glob("*.fab"))

    # Determine the newest .fab file
    newest_fab_file = max(fab_files, key=os.path.getmtime)

    # Run the install command using the newest .fab file
    subprocess.check_call([
        sys.executable,
        "src/patch/flwr_run_patch.py",
        "install",
        str(newest_fab_file)
    ])

def get_dynamic_port():
    """
    Get a dynamically assigned port number.

    Returns:
        int: An available port number assigned by the operating system.
    """
    # Create a socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        # Bind to port 0 to let the OS assign an available port
        s.bind(('127.0.0.1', 0))
        # Get the assigned port number
        port = s.getsockname()[1]
    return port



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""
This file contains utility functions for Secure Aggregation's cipher related
operations.
"""

import random
from typing import Union

import numpy as np
from Crypto.Cipher import AES  # nosec B413
from Crypto.Util.Padding import pad  # nosec B413


def create_ciphertext(
    secret_key: bytes,
    source_id: int,
    destination_id: int,
    seed_share: str,
    key_share: str,
    nonce: bytes = b"nonce",
) -> tuple[bytes, bytes, bytes]:
    """
    Creates a cipher-text using a cipher_key for collaborators source_id and
    destination_id, and share of the private seed and share of the private key.

    The function creates a byte string using the args such that
    data = b'source_id.destination_id.seed_share.key_share'.
    The "." serves as a separator such that all values used to create the
    ciphertext can be easily distinguished when decrypting.

    Args:
        secret_key (bytes): Agreed key in bytes used to construct a cipher for
            the encryption.
        source_id (int): Unique integer ID of the creating collaborator of the
            cipher text.
        destination_id (int): Unique integer ID of the recipient collaborator
            of the cipher text.
        seed_share (bytes): Share of source_id collaborator's private seed for
            destination_id collaborator.
        key_share (bytes): Share of source_id collaborator's private key for
            destination_id collaborator.

    Returns:
        bytes: Ciphertext created using the args.
        bytes: MAC tag for the ciphertext which can be used for verification.
        bytes: Nonce used for generating the cipher used for decryption.
    """
    # Converting the integer collaborator IDs to bytes.
    source_id_bytes = source_id.to_bytes(4, byteorder="big")
    destination_id_bytes = destination_id.to_bytes(4, byteorder="big")
    # Generate the byte string to be encrypted.
    data = (
        source_id_bytes
        + b" "
        + destination_id_bytes
        + b" "
        + str.encode(seed_share)
        + b" "
        + str.encode(key_share)
    )
    # AES cipher requires the secret key to be of a certain length.
    # We use 64 bytes as it is the maximum length available.
    padded_secret_key = pad(secret_key, 64)

    from Crypto.Random import get_random_bytes  # nosec B413

    # Generate a random nonce to make the encryption non-deterministic.
    nonce = get_random_bytes(len(padded_secret_key) // 2)
    # Generate a ciphertext using symmetric block cipher.
    cipher = AES.new(padded_secret_key, AES.MODE_SIV, nonce=nonce)
    ciphertext, mac = cipher.encrypt_and_digest(data)

    return ciphertext, mac, nonce


def decipher_ciphertext(
    secret_key: bytes, ciphertext: bytes, mac: bytes, nonce: bytes
) -> tuple[int, int, bytes, bytes]:
    """
    Decrypt a cipher-text to get the values used to create it.

    The function uses the nonce used while creation of the ciphertext to
    create a cipher. This cipher is used to decypt the ciphertext and verify
    it using the MAC tag, which was also generated during creation of the
    ciphertext.

    Args:
        secret_key (bytes): Agreed key in bytes used to construct a cipher for
            the encryption.
        ciphertext (bytes): Ciphertext to be decrypted.
        mac (bytes): MAC tag for the ciphertext which is used for verification.
        nonce (bytes): Nonce used during cipher generation used for decryption.

    Returns:
        int: Unique integer ID of the creating collaborator of the ciphertext.
        int: Unique integer ID of the recipient collaborator of the ciphertext.
        bytes: Share of source_id collaborator's private seed for
            destination_id collaborator.
        bytes: Share of source_id collaborator's private key for
            destination_id collaborator.
    """
    # Recreate the secret key used for encryption by adding the extra padding.
    padded_secret_key = pad(secret_key, 64)
    # Generate a ciphertext using symmetric block cipher.
    cipher = AES.new(padded_secret_key, AES.MODE_SIV, nonce=nonce)

    data = cipher.decrypt_and_verify(ciphertext, mac)
    # Remove the separator " " from the decrypted data.
    # data = b'source_id destination_id seed_share key_share'
    data = data.split(b" ")

    return (
        # Convert the collaborator IDs to int.
        int.from_bytes(data[0], "big"),
        int.from_bytes(data[1], "big"),
        data[2],
        data[3],
    )


def pseudo_random_generator(seed: Union[int, float, bytes]) -> np.ndarray:
    """
    Generates a random mask using a seed value passed as arg.

    Args:
        seed (Union[int, float, bytes]): Seed to be used for generating a
            pseudo-random number.
        shape (Tuple): Shape of the numpy array to be generated.

    Returns:
        np.ndarray: array with pseudo-randomly generated numbers.
    """
    # Seed random generator.
    random.seed(seed)

    return random.random()


def calculate_shared_mask(agreed_keys: list) -> float:
    """
    Calculate the shared mask based on a list of agreed keys.

    Args:
        agreed_keys (list): A list of tuples where each tuple contains three
            elements:
            - source_index (int): The index of the source.
            - dest_index (int): The index of the destination.
            - agreed_key (Any): The agreed key used for generating the mask.

    Returns:
        float: The total shared mask calculated by adding or subtracting the
            pseudo-random values generated from the agreed keys based on the
            comparison of source and destination indices.
    """
    total_mask = 0.0
    for key in agreed_keys:
        source_index = key[0]
        dest_index = key[1]
        agreed_key = key[2]
        if source_index > dest_index:
            total_mask += pseudo_random_generator(agreed_key)
        elif source_index == dest_index:
            continue
        else:
            total_mask -= pseudo_random_generator(agreed_key)

    return total_mask


def calculate_mask(collaborator_index, agreed_keys, private_seed) -> float:
    """
    Calculates the sum of shared masks between collaborators using
    the agreed keys as the seed for each mask and adding private mask to it.
    """
    total_mask = 0.0
    # Calculating the sum of all shared masks.
    for index in agreed_keys:
        # Using second agreed key to use as seed.
        key = agreed_keys[index][1]
        random.seed(key)
        seed = random.random()
        # mask_u,v = ∆_u,v * PRG(agreed2_u,v) such that
        # ∆_u,v = 1 if u > v else ∆_u,v = -1 if u < v
        # and ∆_u,v = 0 if u = v
        if collaborator_index > index:
            total_mask += seed
        elif collaborator_index < index:
            total_mask -= seed
    # Calculating the private mask for the collaborator.
    random.seed(private_seed)
    total_mask += random.random()

    return total_mask
